from ._logging import logger
from .dataset import Dataset
from . import data, dataset, simulate, ancestry, plot, assoc, estimate, tools, io, cli
from .version import __version__

__all__ = [
    "data",
    "simulate",
    "ancestry",
    "plot",
    "assoc",
    "estimate",
    "tools",
    "io",
    "cli",
    "Dataset",
]



import dask.array as da
import numpy as np
import admix
from typing import List, Tuple, Union, Dict, Sequence
from dask.array import concatenate, from_delayed
from dask.delayed import delayed
from bisect import bisect_left, bisect_right
import dask

is_sorted = lambda a: np.all(a[:-1] <= a[1:])


class Lanc(object):
    """
    Class for local ancestry matrix (n_snp, n_indiv, 2 x ploidy)
    The internal representation is sparse using `breaks` and `values`
    """

    def __init__(
        self,
        path: str = None,
        breaks: List[List[int]] = None,
        values: List[List[str]] = None,
        array: Union[da.Array, np.ndarray] = None,
    ):
        # only one of `path`, `breaks` and `values` can be specified
        assert (path is not None) + ((breaks is not None) & (values is not None)) + (
            array is not None
        ) == 1, "Only one of `path`, `breaks` & `values`, and `array` can be specified"
        if path is not None:
            breaks, values = read_lanc(path)
        elif array is not None:
            breaks, values = array_to_lanc(array)

        assert breaks is not None and values is not None
        n_snp, n_indiv = check_lanc_format(breaks, values)

        self._n_snp = n_snp
        self._n_indiv = n_indiv

        self._breaks = breaks
        self._values = values

    def __repr__(self) -> str:
        descr = (
            f"admix.Lanc object with n_snp x n_indiv = {self.n_snp} x {self.n_indiv}"
        )
        descr += f", use `lanc.dask()` to convert to dask array."

        return descr

    @property
    def n_indiv(self) -> int:
        """Number of individuals."""
        return self._n_indiv

    @property
    def n_snp(self) -> int:
        """Number of SNPs."""
        return self._n_snp

    def lanc_count(self, n_anc: int = None) -> np.ndarray:
        """
        Count the number of local ancestries for each individual.

        Parameters
        ----------
        n_anc : int
            Number of local ancestries in the local ancestries matrix. If None,
            the number of local ancestries is inferred from the data.

        Returns
        -------
            Number of local ancestries for each individual.
        """

        breaks, values = self._breaks, self._values
        n_indiv = self.n_indiv

        # the maximal number of ancestries is assumed to be 10
        MAX_ANC = 10
        lanc_count = np.zeros((n_indiv, MAX_ANC), dtype=int)
        if n_anc is not None:
            lanc_count = lanc_count[:, :n_anc]

        for indiv_i in range(n_indiv):
            start = 0
            for stop, val in zip(breaks[indiv_i], values[indiv_i]):
                a1, a2 = int(val[0]), int(val[1])
                lanc_count[indiv_i, a1] += stop - start
                lanc_count[indiv_i, a2] += stop - start
                start = stop

        # remove any column with zero entries
        if n_anc is None:
            n_anc = np.argmax(np.sum(lanc_count, axis=0) == 0)
        lanc_count = lanc_count[:, :n_anc]
        return lanc_count

    def impute(self, chrom_pos: np.ndarray, dst_chrom_pos: np.ndarray) -> "Lanc":
        """
        Impute missing values in `pos` using `target_pos`.

        Parameters
        ----------
        chrom_pos : np.ndarray
            Chromosomes and positions of the source data.
        dst_chrom_pos
            Chromosomes and positions of the destination data.

        Returns
        -------
        imputed : Lanc
        """
        assert (chrom_pos.shape[1] == 2) and (dst_chrom_pos.shape[1] == 2)
        assert chrom_pos.shape[0] == self.n_snp
        assert is_sorted(chrom_pos[:, 0]) & is_sorted(dst_chrom_pos[:, 0])
        assert set(chrom_pos[:, 0]) == set(
            dst_chrom_pos[:, 0]
        ), "Chromosome set must be the same"
        chrom_list = sorted(set(chrom_pos[:, 0]))

        dst_lanc_list = []
        for chrom in chrom_list:
            src_chrom_idx = np.where(chrom_pos[:, 0] == chrom)[0]
            dst_dst_idx = np.where(dst_chrom_pos[:, 0] == chrom)[0]
            src_chrom_start = src_chrom_idx[0]
            src_chrom_stop = src_chrom_idx[-1] + 1

            chrom_src_breaks, chrom_src_values = lanc_subset_snp_range(
                self._breaks, self._values, src_chrom_start, src_chrom_stop
            )

            chrom_dst_breaks, chrom_dst_values = lanc_impute_single_chrom(
                chrom_src_breaks,
                chrom_src_values,
                chrom_pos[src_chrom_idx, 1],
                dst_chrom_pos[dst_dst_idx, 1],
            )

            dst_lanc_list.append(Lanc(breaks=chrom_dst_breaks, values=chrom_dst_values))
        return concat_lancs(dst_lanc_list)

    def dask(self, snp_chunk: int = 1024):
        return lanc_to_dask(self._breaks, self._values, snp_chunk)

    def numpy(self):
        return lanc_to_numpy(self._breaks, self._values)

    def __getitem__(self, index) -> "Lanc":
        """
        Returns a new instance of `Lanc` with the specified index.
        First index is the SNP, second index is the individual.
        The SNP index can only be a slice with `step` = 1.
        The individual index can be a list of integers or slice, etc.
        """
        if isinstance(index, tuple) and len(index) == 1:
            index = index[0]

        if isinstance(index, tuple):
            if len(index) > 2:
                raise ValueError(
                    "Lanc can only be sliced in SNPs (first dim) and individuals (second dim)"
                )

        from ..dataset._index import unpack_index

        snp_idx, indiv_idx = unpack_index(index)
        # if indiv_idx is slice(None), use all individuals
        if indiv_idx is slice(None):
            indiv_idx = slice(0, self.n_indiv)
        assert isinstance(snp_idx, slice), "SNP index must be a slice"
        assert (snp_idx.step is None) or (
            snp_idx.step is None
        ), "SNP index must with `step` = 1"
        snp_start = snp_idx.start if snp_idx.start is not None else 0
        snp_stop = snp_idx.stop if snp_idx.stop is not None else self.n_snp
        assert snp_start >= 0, "SNP index must start at 0 or higher"
        assert snp_stop <= self.n_snp, "SNP index must stop at n_snp or lower"
        assert snp_start < snp_stop, "SNP index must be a slice with `start < stop`"

        breaks = self._breaks[indiv_idx]
        values = self._values[indiv_idx]
        breaks, values = lanc_subset_snp_range(breaks, values, snp_start, snp_stop)
        return Lanc(breaks=breaks, values=values)

    def write(
        self,
        path: str,
    ):
        breaks, values = self._breaks, self._values
        n_snp, n_indiv = check_lanc_format(breaks, values)

        # write to file
        lines = []

        # header
        lines.append(f"{n_snp} {n_indiv}")

        for indiv_break, indiv_value in zip(breaks, values):
            lines.append(
                " ".join([str(b) + ":" + v for (b, v) in zip(indiv_break, indiv_value)])
            )

        with open(path, "w") as f:
            f.writelines("\n".join(lines))


def assign_lanc(dset: admix.Dataset, lanc_file: str, format: str = "rfmix"):
    """
    Assign local ancestry to a dataset. Currently we assume that the rfmix file contains
    2-way admixture information.

    Parameters
    ----------
    dset: admix.Dataset
        Dataset to assign local ancestry to.
    lanc_file: str
        Path to local ancestry data.
    format: str
        Format of local ancestry data.
        Currently only "rfmix" is supported.

    Returns
    -------
    dset: admix.Dataset
        Dataset with local ancestry assigned.
    TODO:
    - Add support for other formats.
    """
    import pandas as pd
    import numpy as np

    assert format in ["rfmix"], "Only rfmix format is supported."
    # assign local ancestry
    rfmix = pd.read_csv(lanc_file, sep="\t", skiprows=1)

    lanc_full = da.full(
        shape=(dset.dims["indiv"], dset.dims["snp"], dset.dims["ploidy"]),
        fill_value=-1,
        dtype=np.int8,
    )
    lanc0 = rfmix.loc[:, rfmix.columns.str.endswith(".0")].rename(
        columns=lambda x: x[:-2]
    )
    lanc1 = rfmix.loc[:, rfmix.columns.str.endswith(".1")].rename(
        columns=lambda x: x[:-2]
    )

    assert np.all(dset.indiv == lanc0.columns)
    assert np.all(dset.indiv == lanc1.columns)

    for i_row, row in rfmix.iterrows():
        mask_row = (
            (row.spos <= dset.snp["POS"]) & (dset.snp["POS"] <= row.epos)
        ).values
        lanc_full[:, mask_row, 0] = lanc0.loc[i_row, :].values[:, np.newaxis]
        lanc_full[:, mask_row, 1] = lanc1.loc[i_row, :].values[:, np.newaxis]

    dset_names = tuple(d for d in dset.dims)
    if dset_names == ("indiv", "snp", "ploidy"):
        # do nothing
        pass
    elif dset_names == ("snp", "indiv", "ploidy"):
        lanc_full = lanc_full.swapaxes(0, 1)
    else:
        raise ValueError(
            f"Unexpected dimensions {dset_names}. "
            "Expected (indiv, snp, ploidy) or (snp, indiv, ploidy)"
        )

    lanc_full = lanc_full.rechunk(dset.geno.chunks)

    dset = dset.assign({"lanc": (dset_names, lanc_full)})
    dset = dset.assign_attrs({"n_anc": 2})
    return dset


def find_closest_index(a, x, tie="left"):
    """
    Locate the index in a of the closest value to x.
    When two values are equally close, return the smallest index.

    Parameters
    ----------
    a: array-like
        Values to search.
    x: float
        Value to search for.
    tie: str
        How to handle ties. left -> return the smallest index.
        right -> return the largest index. default: left
    """
    assert tie in ["left", "right"], "tie must be 'left' or 'right'"
    if tie == "left":
        op = np.greater_equal
    else:
        op = np.greater

    idx = bisect_left(a, x)

    if idx == len(a):
        # x is larger than a[-1]
        idx = len(a) - 1
    elif idx > 0 and op(a[idx] - x, x - a[idx - 1]):
        # x is closer to a[idx-1] than a[idx]
        idx -= 1
    return idx


def read_bp_lanc(path: str):
    """
    read .bp file (breakpoints file) and convert to lanc object.
    The format is specified at
    https://github.com/williamslab/admix-simu#output-breakpoints-file-out_prefixbp

    Parameters
    ----------
    path : str
        path to .bp file

    Returns
    -------
    admix.data.Lanc
        the converted local ancestry

    """
    assert path.endswith(".bp")
    with open(path, "r") as f:
        bp_data = [line.strip().split() for line in f.readlines()][1:]
    hap_breaks = [[int(l.split(":")[1]) + 1 for l in line] for line in bp_data]
    hap_values = [[l.split(":")[0] for l in line] for line in bp_data]

    dip_breaks, dip_values = clean_lanc(
        *admix.data.haplo2diplo(breaks=hap_breaks, values=hap_values),
        remove_repeated_val=True,
    )
    lanc = admix.data.Lanc(breaks=dip_breaks, values=dip_values)
    return lanc


def concat_lancs(lanc_list: List[Lanc], dim="snp") -> Lanc:
    """Concatenate local ancestry files
    For example (1:01, 15: 00) and (3:01, 12:01) will be merged to (1:01, 15:00, 18:01, 27:01)
    Parameters
    ----------
    lanc_list: List[Lanc]
        List of local ancestry files.

    Returns
    -------
    """
    assert dim == "snp", "Only concatenating along snp is supported."

    snp_start = 0
    n_indiv = lanc_list[0].n_indiv
    breaks: List[List[int]] = [[] for _ in range(n_indiv)]
    values: List[List[str]] = [[] for _ in range(n_indiv)]
    for lanc in lanc_list:
        assert lanc.n_indiv == n_indiv, "Number of individuals must be the same."
        for indiv_i in range(n_indiv):
            breaks[indiv_i].extend([b + snp_start for b in lanc._breaks[indiv_i]])
            values[indiv_i].extend(lanc._values[indiv_i])
        snp_start += lanc.n_snp
    return Lanc(breaks=breaks, values=values)


def read_lanc(path: str) -> Tuple[List, List]:
    """Read local ancestry with .lanc format

    Parameters
    ----------
    path : .lanc file
        Local ancestry file

    Returns
    -------
    da.Array
        (n_snp, n_indiv, 2)
    """

    # TODO: first check input
    # the end must equal to n_snp
    with open(path) as f:
        lines = f.readlines()
    n_snp0, n_indiv0 = [int(i) for i in lines[0].strip().split()]
    data_list = [line.strip().split() for line in lines[1:]]
    assert len(data_list) == n_indiv0
    breaks = [[int(l.split(":")[0]) for l in line] for line in data_list]
    values = [[l.split(":")[1] for l in line] for line in data_list]

    n_snp, n_indiv = check_lanc_format(breaks, values)
    assert n_snp0 == n_snp, "n_snp in .lanc file does not match"
    assert n_indiv0 == n_indiv, "n_indiv in .lanc file does not match"
    return breaks, values


def check_lanc_format(breaks: List[List[int]], values: List[List[str]]):
    """Check the format of .lanc file

    Parameters
    ----------
    breaks : List[List[int]]
        List of break points
    values : List[List[str]]
        List of value

    Returns
    -------
    Tuple[int, int]
        (n_snp, n_indiv)
    """
    assert len(breaks) == len(
        values
    ), "`breaks` and `values` must have the same length (same as n_indiv)"

    assert np.all([len(b) == len(v) for b, v in zip(breaks, values)])
    n_snp = breaks[0][-1]
    n_indiv = len(breaks)
    assert np.all(
        [n_snp == b[-1] for b in breaks]
    ), "The last element of break points for each individual in `break_list` must be equal"
    return n_snp, n_indiv


def clean_lanc(
    breaks: List[List[int]], values: List[List[str]], remove_repeated_val: bool = False
):
    """Clean up local ancestry file

    Parameters
    ----------
    breaks : List[List[int]]
        break points
    values : List[List[str]]
        values
    remove_same_val : bool
        Remove segments with same values (default: False)
        For example, 50:01 100:01 300:00 -> 100:01 300:00
        For example, 50:01 100:01 100:10 300:00 300:01 -> 100:01 300:00

    """
    new_breaks = []
    new_values = []

    if not remove_repeated_val:
        for br, vl in zip(breaks, values):
            # remove duplicated break positions, only preserve the first one
            d = dict()
            for b, v in zip(br, vl):
                if b not in d:
                    d[b] = v

            d.pop(0, None)
            new_breaks.append(list(d.keys()))
            new_values.append(list(d.values()))
    else:
        for br, vl in zip(breaks, values):
            # remove duplicated break positions, only preserve the first one
            last_break = None
            last_value = None
            d = dict()
            for b, v in zip(br, vl):
                if (last_value is not None) and (v != last_value):
                    # this value is different from the last one
                    # record last one as break points
                    if last_break not in d:
                        d[last_break] = last_value
                # either: last_value is None
                # or: this value is the same as the last one, extend
                last_break, last_value = b, v

            # record finel one
            assert last_break is not None
            if last_break not in d:
                d[last_break] = last_value

            d.pop(0, None)
            new_breaks.append(list(d.keys()))
            new_values.append(list(d.values()))
    return new_breaks, new_values


def haplo2diplo(breaks: List[List], values: List[List]):
    """convert haplotype to diplotype, combine every 2 haplotypes into 1 diplotype

    Parameters
    ----------
    breaks : List[List[int]]
        break points
    values : List[List[str]]
        values

    Returns
    -------
    List[List[int]]
        break points
    List[List[str]]
        values
    """
    assert len(breaks) == len(values)
    n_haplo = len(breaks)
    assert n_haplo % 2 == 0, "Number of haplotypes must be even"
    n_diplo = n_haplo // 2

    new_breaks = []
    new_values = []
    for i in range(n_diplo):
        br1, br2 = breaks[i * 2 : (i + 1) * 2]
        vl1, vl2 = values[i * 2 : (i + 1) * 2]
        unique_br = np.union1d(br1, br2).tolist()
        new_vl = [
            str(v1) + str(v2)
            for v1, v2 in zip(
                [vl1[bisect_left(br1, b)] for b in unique_br],
                [vl2[bisect_left(br2, b)] for b in unique_br],
            )
        ]
        new_breaks.append(unique_br)
        new_values.append(new_vl)
    return new_breaks, new_values


def lanc_subset_snp_range(
    breaks: List[List[int]], values: List[List[str]], start: int, stop: int
):
    """
    Subset the .lanc file

    Parameters
    ----------
    start : int
        start of SNP
    stop : int
        stop of SNP
    break_list: List[List[int]]
        list of break points
    value_list: List[List[str]]
    """
    # TODO: double check this function whether usage of bisect_left is correct
    # For each individual, find index of break points that's within [start, stop]
    start_idx = [bisect_left(indiv_break, start) for indiv_break in breaks]
    stop_idx = [bisect_left(indiv_break, stop) for indiv_break in breaks]
    new_breaks = [br[s:e] + [stop] for s, e, br in zip(start_idx, stop_idx, breaks)]
    # offset with start
    new_breaks = [[b - start for b in br] for br in new_breaks]
    # find corresponding value
    new_values = [val[s:e] + [val[e]] for s, e, val in zip(start_idx, stop_idx, values)]

    # clean up
    new_breaks, new_values = clean_lanc(new_breaks, new_values)
    return new_breaks, new_values


def lanc_to_dask(
    breaks: List[List[int]], values: List[List[str]], snp_chunk: int = 1024
):
    """
    Given `break_list` list of break points with `n_indiv` length
    And the corresponding `value_list`, the correponding value

    Convert to dask matrix
    """
    n_snp, n_indiv = check_lanc_format(breaks, values)

    # all local ancestries
    lancs = []

    subset_dense = lambda start, stop: lanc_to_numpy(
        *lanc_subset_snp_range(breaks, values, start, stop)
    )
    snp_start = 0
    while snp_start < n_snp:
        snp_stop = min(snp_start + snp_chunk, n_snp)
        shape = (snp_stop - snp_start, n_indiv, 2)

        lancs.append(
            from_delayed(
                value=delayed(subset_dense)(snp_start, snp_stop),
                shape=shape,
                dtype=np.int8,
            )
        )
        snp_start = snp_stop
    return concatenate(lancs, 0, False)


def lanc_to_numpy(breaks, values):
    """
    Given `break_list` list of break points with `n_indiv` length
    And the corresponding `value_list`, the correponding value

    Convert to dense matrix
    """
    n_snp, n_indiv = check_lanc_format(breaks, values)
    mat = np.full((n_snp, n_indiv, 2), -1)
    for indiv_i in range(n_indiv):
        start = 0
        for stop, val in zip(breaks[indiv_i], values[indiv_i]):
            a1, a2 = int(val[0]), int(val[1])
            mat[start:stop, indiv_i, 0] = a1
            mat[start:stop, indiv_i, 1] = a2
            start = stop
    return mat


def array_to_lanc(array: Union[da.Array, np.ndarray]) -> Tuple[List, List]:
    """Convert dense array to local ancestry format

    Parameters
    ----------
    array : np.ndarray
        (n_snp, n_indiv, 2)

    Returns
    """
    n_snp, n_indiv, n_ploidy = array.shape
    assert n_ploidy == 2, "`lanc` must be (n_snp, n_indiv, 2)"

    # convert to dask.array if numpy array
    if isinstance(array, np.ndarray):
        lanc = da.from_array(array)
    else:
        lanc = array

    assert isinstance(lanc, da.Array), "`lanc` must be dask array"

    # switch points
    snp_pos, indiv_pos, _, = dask.compute(
        np.where(lanc[1:, :, :] != lanc[0:-1, :, :]),
        scheduler="single-threaded",
    )[0]
    # end points for all the individuals
    snp_pos = np.concatenate([snp_pos, [n_snp - 1] * n_indiv])
    indiv_pos = np.concatenate([indiv_pos, np.arange(n_indiv)])
    # (snp, indiv) -> snp * n_indiv + indiv
    values = lanc.reshape([-1, 2])[indiv_pos + snp_pos * n_indiv, :].compute()
    values = np.array([str(v[0]) + str(v[1]) for v in values])

    break_list = []
    value_list = []
    for indiv_i in range(n_indiv):
        indiv_mask = indiv_pos == indiv_i
        # +1 because .lanc denote the [start, stop) right-open interval
        indiv_snp_pos, unique_mask = np.unique(
            snp_pos[indiv_mask] + 1, return_index=True
        )
        indiv_values = values[indiv_mask][unique_mask]
        break_list.append(indiv_snp_pos.tolist())
        value_list.append(indiv_values.tolist())
    return break_list, value_list


def lanc_impute_single_chrom(
    src_breaks: List[List[int]],
    src_values: List[List[str]],
    src_pos: np.ndarray,
    dst_pos: np.ndarray,
):
    """local ancestry imputation for single chromosome

    Using the following steps:
        1. basic checks are performed for the two data sets.
        2. `dset_ref`'s individuals is matched with `dset`, `dset`'s individuals
            must be a subset of `dset_ref`'s individuals.
        3. Imputation is performed based on that, for each position in `dst`, find the
        nearest SNP in `src` measured by physical distance, and use the assignment in src.
        if there is a tie, use the one in `src` with the smallest index.

    Parameters
    ----------
    ref_pos : np.ndarray
        position in reference data sets
    ref_lanc : admix.data.Lanc
        reference local ancestry
    sample_pos : np.ndarray
        position in sample data sets to be imputed
    """
    src_n_snp, n_indiv = check_lanc_format(src_breaks, src_values)
    dst_n_snp = len(dst_pos)
    assert len(src_pos) == src_n_snp
    assert is_sorted(src_pos) & is_sorted(dst_pos), "src_pos and dst_pos must be sorted"
    assert (src_pos.ndim == 1) & (
        dst_pos.ndim == 1
    ), "src_pos and dst_pos must be 1-D arrays"
    assert np.issubdtype(src_pos.dtype, np.integer) & np.issubdtype(
        dst_pos.dtype, np.integer
    ), "pos and dst_pos must be integer type"

    # find physical position of break points in src data
    # `b - 1` because breaks corresponds to right-open interval

    dst_breaks = []
    dst_values = []
    for indiv_src_breaks, indiv_src_values in zip(src_breaks, src_values):
        # find a set of candidate index for dst break points
        # for each break points in src_pos, define `mid`` = (src_pos[b] + src_pos[b+1])/2
        # then find the closest point in dst that's smaller or equal to `mid`
        # then assign the local ancestry at b in src to that point
        # max(b - 1, 0) is to avoid accidental break = 0 (which should not appear)
        # we threshold by 0 anyway to avoid negative index
        indiv_dst_breaks = [
            bisect_right(
                dst_pos, (src_pos[max(b - 1, 0)] + src_pos[b]) / 2  # type: ignore
            )
            for b in indiv_src_breaks[:-1]
        ] + [dst_n_snp]
        dst_breaks.append(indiv_dst_breaks)
        dst_values.append(indiv_src_values)

    return clean_lanc(dst_breaks, dst_values)


###############################################################################
# LEGACY code for impute_lanc
###############################################################################

# version 2
# to be computational efficient, we start with src data
# for each src break point, find the closest dst break point
# dst_breaks = []
# dst_values = []
# for indiv_src_break_pos, indiv_src_values in zip(src_break_pos, src_values):
#     # find a set of candidate index for dst break points
#     indiv_dst_breaks = [
#         find_closest_index(dst_pos, p, tie="right") + 1
#         for p in indiv_src_break_pos[:-1]
#     ] + [dst_n_snp]
#     # assign src values to dst break points
#     indiv_dst_values = [
#         indiv_src_values[
#             find_closest_index(indiv_src_break_pos, dst_pos[i - 1], tie="left")
#         ]
#         for i in indiv_dst_breaks
#     ]

#     dst_breaks.append(indiv_dst_breaks)
#     dst_values.append(indiv_dst_values)

# version 1
# dict_tmp: Dict[int, Tuple] = {}
# for p, v in zip(indiv_src_break_pos[:-1], indiv_src_values[:-1]):
#     # tie is right such that for the dst, tie is always the one with the smallest index
#     idx = find_closest_index(dst_pos, p, tie="right")
#     dist = np.abs(dst_pos[idx] - p)
#     print(f"src_break_pos: {p}, src_value: {v}, dst_idx: {idx}, dist: {dist}")
#     # replace the dist, value if the new distance is smaller
#     if (idx not in dict_tmp) or (dist < dict_tmp[idx][0]):
#         dict_tmp[idx] = (dist, v)
# # the last src break point, always correspond to the last dst break point
# dst_breaks.append([k + 1 for k in dict_tmp.keys()] + [dst_n_snp])
# dst_values.append([v[1] for v in dict_tmp.values()] + [indiv_src_values[-1]])


# def impute_lanc(dset: admix.Dataset, dset_ref: admix.Dataset):
#     """
#     Impute local ancestry using a reference dataset. The two data sets are assumed to
#     have the same haplotype order, etc. Typically they are just a subset of each other.

#     Using the following steps:
#         1. basic checks are performed for the two data sets.
#         2. `dset_ref`'s individuals is matched with `dset`, `dset`'s individuals
#             must be a subset of `dset_ref`'s individuals.
#         3. Imputation are performed

#     Parameters
#     ----------
#     dset: a data set to be imputed with local ancestry
#     dset_ref: a data set with local ancestry for reference

#     Returns
#     -------
#     dset_imputed: a data set with imputed local ancestry
#     """
#     assert (
#         len(set(dset.snp.CHROM)) == 1
#     ), "Data set to be imputed can only have one chromosome"

#     assert (
#         len(set(dset_ref.snp.CHROM)) == 1
#     ), "Data set to be imputed can only have one chromosome"

#     # dset.indiv is a subset of dset_ref.indiv
#     assert np.all(
#         dset.indiv.index, dset_ref.indiv.index
#     ), "dset and dset_ref must have the same individuals"

#     # find relevant regions in reference dataset with local ancestry (hapmap3 SNPs here)
#     ref_start = np.argmin(np.abs(dset_ref.snp.POS.values - dset.snp.POS.values[0]))
#     ref_stop = np.argmin(np.abs(dset_ref.snp.POS.values - dset.snp.POS.values[-1]))

#     # subset reference dataset to relevant regions
#     dset_ref = dset_ref[ref_start : ref_stop + 1]

#     imputed_lanc = []
#     for ploidy_i in range(2):
#         # form a dataframe which contains the known local ancestry and locations to be imputed
#         df_snp = pd.concat([dset.snp[["POS"]], dset_ref.snp[["POS"]]])
#         df_snp = df_snp[~df_snp.index.duplicated()].sort_values("POS")

#         df_lanc = pd.DataFrame(
#             index=df_snp.index.values, columns=dset.indiv.index.values, dtype=float
#         )
#         # fill margin
#         df_lanc.iloc[0, :] = dset_ref.lanc[0, :, ploidy_i]
#         df_lanc.iloc[-1, :] = dset_ref.lanc[-1, :, ploidy_i]

#         # fill inside
#         df_lanc.loc[dset_ref.snp.values, :] = dset_ref["lanc"][:, :, ploidy_i].values

#         # interpolate
#         df_lanc = df_lanc.reset_index().interpolate(method="nearest").set_index("index")

#         imputed_lanc.append(df_lanc.loc[dset["snp"].values, :].values.astype(np.int8))

#     # imputed_lanc is in the order of ("snp", "indiv", "ploidy")
#     return da.from_array(np.dstack(imputed_lanc), chunks=-1)



import subprocess
import tempfile
from os.path import join
from ..utils import get_cache_dir, cd
import urllib.request
import shutil
import os
import admix


def get_dependency(name, download=True):
    """Get path to an depenency
    Find the binary in the following locations:
    - $PATH
    - package installment directory admix-tools/.admix_cache/bin/<name>

    If not found in any of these locations, download the corresponding software package
    - plink: https://www.cog-genomics.org/plink/2.0/
    - gcta: https://github.com/gcta/gcta

    Parameters
    ----------
    download : bool
        whether to download plink if not found

    Returns
    -------
    Path to binary executable
    """
    # find in path
    if shutil.which(name):
        return shutil.which(name)

    # find in cache
    cache_dir = join(get_cache_dir(), "bin")
    os.makedirs(cache_dir, exist_ok=True)
    cache_bin_path = join(cache_dir, name)
    if os.path.exists(cache_bin_path):
        return cache_bin_path
    else:
        # download
        if download:
            from sys import platform

            if name == "plink2":
                if platform == "darwin":
                    url = "https://s3.amazonaws.com/plink2-assets/alpha2/plink2_mac.zip"
                elif platform == "linux":
                    url = "https://s3.amazonaws.com/plink2-assets/alpha2/plink2_linux_x86_64.zip"
                else:
                    raise ValueError(f"Unsupported platform {platform}")

                with tempfile.TemporaryDirectory() as tmp_dir:
                    with cd(tmp_dir):
                        urllib.request.urlretrieve(
                            url,
                            "file.zip",
                        )
                        subprocess.check_call(f"unzip file.zip -d dir", shell=True)
                        subprocess.check_call(
                            f"mv dir/plink2 {cache_bin_path}", shell=True
                        )

            elif name == "plink":
                if platform == "darwin":
                    url = (
                        "https://s3.amazonaws.com/plink1-assets/plink_mac_20210606.zip"
                    )
                elif platform == "linux":
                    url = "https://s3.amazonaws.com/plink1-assets/plink_linux_x86_64_20210606.zip"
                else:
                    raise ValueError(f"Unsupported platform {platform}")

                with tempfile.TemporaryDirectory() as tmp_dir:
                    with cd(tmp_dir):
                        urllib.request.urlretrieve(
                            url,
                            "file.zip",
                        )
                        subprocess.check_call(f"unzip file.zip -d dir", shell=True)
                        subprocess.check_call(
                            f"mv dir/plink {cache_bin_path}", shell=True
                        )

            elif name == "gcta64":
                if platform == "darwin":
                    platform_wildcard = "gcta_1.93.2beta_mac"
                elif platform == "linux":
                    platform_wildcard = "gcta_1.93.2beta"
                else:
                    raise ValueError(f"Unsupported platform {platform}")
                url = (
                    f"https://cnsgenomics.com/software/gcta/bin/{platform_wildcard}.zip"
                )
                with tempfile.TemporaryDirectory() as tmp_dir:
                    with cd(tmp_dir):
                        urllib.request.urlretrieve(
                            url,
                            "file.zip",
                        )
                        subprocess.check_call(f"unzip file.zip -d dir", shell=True)
                        subprocess.check_call(
                            f"mv dir/{platform_wildcard}/gcta64 {cache_bin_path}",
                            shell=True,
                        )

            elif name == "liftOver":
                if platform == "darwin":
                    url = "http://hgdownload.soe.ucsc.edu/admin/exe/macOSX.x86_64/liftOver"
                # NOTE: in case GLIBC version is not compatible, try
                # http://hgdownload.soe.ucsc.edu/admin/exe/linux.x86_64.v369/liftOver
                elif platform == "linux":
                    url = (
                        "http://hgdownload.soe.ucsc.edu/admin/exe/linux.x86_64/liftOver"
                    )
                else:
                    raise ValueError(f"Unsupported platform {platform}")

                with tempfile.TemporaryDirectory() as tmp_dir:
                    with cd(tmp_dir):
                        urllib.request.urlretrieve(
                            url,
                            "liftOver",
                        )
                        subprocess.check_call(f"chmod +x liftOver", shell=True)
                        subprocess.check_call(
                            f"mv liftOver {cache_bin_path}", shell=True
                        )

            elif name == "hapgen2":
                if platform == "linux":
                    url = (
                        "http://mathgen.stats.ox.ac.uk/genetics_software/hapgen"
                        "/download/builds/x86_64/v2.2.0/hapgen2_x86_64.tar.gz"
                    )
                else:
                    raise ValueError(f"Unsupported platform {platform}")
                with tempfile.TemporaryDirectory() as tmp_dir:
                    with cd(tmp_dir):
                        urllib.request.urlretrieve(
                            url,
                            "file.tar.gz",
                        )
                        subprocess.check_call(f"tar -xvf file.tar.gz", shell=True)

                        subprocess.check_call(
                            f"mv hapgen2 {cache_bin_path}",
                            shell=True,
                        )

            elif name == "admix-simu":
                if platform == "linux":
                    url = "https://github.com/williamslab/admix-simu/archive/refs/heads/master.zip"
                else:
                    raise ValueError(f"Unsupported platform {platform}")
                with tempfile.TemporaryDirectory() as tmp_dir:
                    with cd(tmp_dir):
                        urllib.request.urlretrieve(
                            url,
                            "master.zip",
                        )
                        subprocess.check_call(f"unzip master.zip -d dir", shell=True)
                        # make
                        subprocess.check_call(
                            f"cd dir/admix-simu-master && make", shell=True
                        )
                        subprocess.check_call(
                            f"mv dir/admix-simu-master/ {cache_bin_path}",
                            shell=True,
                        )
            else:
                raise ValueError(f"Unsupported software {name}")

        else:
            raise ValueError(
                f"{name} not found in $PATH or {cache_dir}, set `download=True` to download from website"
            )

        return cache_bin_path


def get_cache_data(name: str, **kwargs) -> str:
    """
    Obtain the path to the cached data.

    Find the data in the following locations:
    - package installation directory admix-tools/.admix_cache/data/<name>

    If not found in any of these locations, download will start

    Parameters
    ----------
    name: str
        name of the data
        - genetic_map: used for HAPGEN2, download from
            https://alkesgroup.broadinstitute.org/Eagle/downloads/tables/
            kwargs["build"] = hg19 or hg38
        - hapmap3_snps: used to intersect with HM3 SNPs, download from
            https://ndownloader.figshare.com/files/25503788 (see https://privefl.github.io/bigsnpr/articles/LDpred2.html)

    Returns
    -------
    str: path to the cached file
    """

    # find in cache
    cache_dir = join(get_cache_dir(), "data", name)
    os.makedirs(cache_dir, exist_ok=True)

    if name == "genetic_map":
        assert kwargs["build"] in ["hg19", "hg38"]
        file_name = f"genetic_map_{kwargs['build']}_withX.txt.gz"
        cache_path = join(cache_dir, file_name)
        url = (
            "https://storage.googleapis.com/broad-alkesgroup-public/Eagle/downloads/tables/"
            + file_name
        )
    elif name == "hapmap3_snps":
        file_name = "hapmap3_snps.rds"
        cache_path = join(cache_dir, file_name)
        url = "https://ndownloader.figshare.com/files/25503788"
    else:
        raise ValueError(f"Unsupported data {name}")

    if not os.path.exists(cache_path):
        admix.logger.info(f"{name} not found at {cache_path}.")
        admix.logger.info(f"Downloading {name} from {url}.")
        admix.logger.info(
            f"If this gets stuck or fails, manually download {url} to {cache_path}."
        )

        try:
            urllib.request.urlretrieve(
                url,
                cache_path,
            )
        except Exception as e:
            admix.logger.info(
                f"Download failed. Please manually download {url} to {cache_path}."
            )
            raise e
    else:
        admix.logger.info(f"{name} found at {cache_path}.")
    return cache_path



from ._read import (
    read_digit_mat,
    read_vcf,
    read_dataset,
    read_lanc,
    read_gcta_grm,
    read_rfmix,
    read_joint_pca,
)
from ._write import write_gcta_grm, write_digit_mat, write_lanc, write_dataset


from typing import List
from ._dataset import Dataset, is_aligned
from ._load import (
    load_toy,
    load_toy_admix,
    load_ukb_eur_afr_hm3,
    get_test_data_dir,
    download_simulated_example_data,
)

# required columns in dset.snp
REQUIRED_SNP_COLUMNS = ["CHROM", "POS", "REF", "ALT"]



import numpy as np
from scipy import linalg
import pandas as pd
import xarray as xr
from typing import List, Union
import admix
import tempfile
from ..utils import cd
import dask


def variance_component(
    dset: admix.Dataset,
    grm: Union[str, List[str], dict],
    pheno: np.ndarray,
    cov_cols: List[str] = None,
    cov_intercept: bool = True,
    method: str = "HE",
):
    """
    Common interface for fitting variance components.
    Including HE-regression and REML.

    Parameters
    ----------
    dset: admix.Dataset
        Dataset to estimate correlation from.
    grm: str, list of str or dict
        column name(s) of GRM, or a dict of {name: grm}
        Don't include the identity matrix, the indentify matrix representing environemntal
        factor will be added to the list automatically
    pheno: np.ndarray
        Phenotypes to estimate genetic correlation. If a matrix is provided, then each
        column is treated as a separate phenotype.
    cov_cols: list, optional
        List of covariate columns.
    cov_intercept: bool, optional
        Whether to include intercept in covariate matrix.
    method: str, optional
        Method to use for estimation. Valid options are:
            - "HE" (default): Haseman–Elston (HE) regression.
            - "RHE": Randomized Haseman–Elston (RHE) regression.
            - "REML": REML regression.
            - "HE-gcta": HE regression as implemented in GCTA.
            - "REML-gcta": REML as implemented in GCTA.

    """

    assert method in ["HE", "HE-gcta", "RHE", "REML"]
    n_indiv = dset.dims["indiv"]
    # get the dictionary of all the GRMs (including environmental)
    # name -> grm
    if isinstance(grm, dict):
        # obtain from the list
        grm_dict = {k: grm[k] for k in grm}
    elif isinstance(grm, list):
        grm_dict = {k: dset[k].data for k in grm}
    elif isinstance(grm, str):
        grm_dict = {grm: dset[grm].data}
    else:
        raise ValueError("`grm` must be a dictionary, list or string")

    n_indiv = dset.dims["indiv"]

    # build the covariate matrix
    if cov_cols is not None:
        cov_values = np.zeros((n_indiv, len(cov_cols)))
        for i_cov, cov_col in enumerate(cov_cols):
            cov_values[:, i_cov] = dset[cov_col + "@indiv"].values
        if cov_intercept:
            cov_values = np.c_[np.ones((n_indiv, 1)), cov_values]
    else:
        cov_values = None
        if cov_intercept:
            cov_values = np.ones((n_indiv, 1))

    if pheno.ndim == 1:
        pheno = pheno.reshape((-1, 1))
    assert pheno.shape[0] == n_indiv

    n_pheno = pheno.shape[1]

    if method == "HE" or method == "HE-gcta":

        rls_list = HE_reg(
            grm_list=[grm_dict[name] for name in grm_dict],
            pheno=pheno,
            cov=cov_values,
            method="built-in" if method == "HE" else "gcta",
        )
        if method == "HE":
            columns = [name for name in grm_dict] + ["e"]
        elif method == "HE-gcta":
            columns = [name + "/total" for name in grm_dict]
        else:
            raise ValueError("Invalid method")
        df_rls = pd.DataFrame(rls_list, columns=columns)

    elif method == "REML":
        rls_list = REML(
            grm_list=[grm_dict[name] for name in grm_dict], pheno=pheno, cov=cov_values
        )
    else:
        raise NotImplementedError("{} is not implemented".format(method))

    return df_rls


def HE_reg(
    grm_list: List,
    pheno: np.ndarray,
    cov: np.ndarray = None,
    method="built-in",
):
    """
    Estimate the variance components for the given phenotype with the
    Haseman-Elston (HE) regression. Note that the enviornmental GRM (identity matrix)
    should be not included in the list of GRMs.

    Parameters
    ----------
    grm_list : list of array-like
        List of GRMs to be used.
    pheno : array-like
        Phenotypes to estimate genetic correlation. If a matrix is provided, then each
        column is treated as a separate phenotype.
    cov : array-like, optional
        Covariates (n_indiv, n_cov) to be used in the regression. If None, then no covariates are used.
    method : str, optional
        Method to use.
            "built-in": built-in implementation in admix.estimators.HE_reg
            "gcta": use GCTA to estimate the variance components
    Returns
    -------
    df_rls : pandas.DataFrame
        DataFrame with the regression coefficients.

    """

    quad_form = lambda x, A: np.dot(np.dot(x.T, A), x)

    n_indiv = grm_list[0].shape[0]
    grm_list = grm_list + [np.eye(n_indiv)]
    n_grm = len(grm_list)
    n_pheno = pheno.shape[1]

    # list of results (one for each simulation)
    rls_list = []
    assert method in [
        "built-in",
        "gcta",
    ], "`method` must be one of: ['built-in', 'gcta']"

    # build projection matrix from covariate matrix
    if cov is None:
        cov_proj_mat = np.eye(n_indiv)
    else:
        # TODO: check colinearity of the covariates?
        cov_proj_mat = np.eye(n_indiv) - np.linalg.multi_dot(
            [
                cov,
                np.linalg.inv(np.dot(cov.T, cov)),
                cov.T,
            ]
        )
    # project the grm and the phenotypes to the new space
    grm_list = [np.dot(grm, cov_proj_mat) for grm in grm_list]
    pheno = np.dot(cov_proj_mat, pheno)

    if method == "built-in":
        HE_design = np.zeros((n_grm, n_grm))

        # fill in `HE_design`
        for i in range(n_grm):
            for j in range(n_grm):
                if i <= j:
                    HE_design[i, j] = (grm_list[i] * grm_list[j]).sum()
                    HE_design[j, i] = HE_design[i, j]

        # build response vector for each phenotype
        for i_pheno in range(n_pheno):
            HE_response = np.zeros(n_grm)
            for i in range(n_grm):
                HE_response[i] = quad_form(pheno[:, i_pheno], grm_list[i])
            rls_list.append(linalg.solve(HE_design, HE_response))

    elif method == "gcta":
        # creating dummy individual IDs
        df_id = pd.DataFrame(
            {
                "FID": np.arange(n_indiv).astype(str),
                "IID": np.arange(n_indiv).astype(str),
            }
        )

        # create temporary directory
        with tempfile.TemporaryDirectory() as tmp_dir:
            print(tmp_dir)
            with cd(tmp_dir):
                f = open("mgrm.txt", "w")
                # the last grm corresponds to identity matrix, which will be coped
                # internally by GCTA
                for i_grm, grm in enumerate(grm_list[:-1]):
                    name = f"grm_{i_grm}"
                    # NOTE: the degree of freedom for GCTA implementation is not
                    # properly set up. Should not be a big problem for the current.

                    # NOTE: n_snps will not be used by GCTA in this HE regression.
                    # so we just set it to some arbitrary value `1`.
                    admix.io.write_gcta_grm(
                        file_prefix=name,
                        grm=dask.compute(np.dot(grm, cov_proj_mat))[0],
                        df_id=df_id,
                        n_snps=np.array([1] * n_indiv),
                    )
                    f.write(f"{name}\n")
                f.close()
                df_pheno = df_id.copy()
                for i_pheno in range(n_pheno):
                    df_pheno["PHENO"] = np.dot(cov_proj_mat, pheno[:, i_pheno])
                    df_pheno.to_csv("pheno.txt", sep="\t", index=False, header=False)

                    # run GCTA
                    admix.tools.gcta(
                        "--HEreg --mgrm mgrm.txt --pheno pheno.txt --out result"
                    )
                    df_rls = pd.read_csv(
                        "result.HEreg",
                        delim_whitespace=True,
                        skiprows=1,
                        nrows=n_grm,
                    )
                    rls_list.append(df_rls["Estimate"][1:].values)

    else:
        raise NotImplementedError

    return rls_list


def REML(grm_list: List, pheno: np.ndarray, cov: np.ndarray):
    """
    Estimate the variance components for the given phenotype with REML.
    Currently only a GCTA wrapper is provided.
    Parameters
    ----------
    grm_list : list of array-like
        List of GRMs to be used.
    pheno : array-like
        Phenotypes to estimate genetic correlation. If a matrix is provided, then each
        column is treated as a separate phenotype.
    cov : array-like
        Covariates to be used. (n_indiv, n_cov)
    """
    # creating dummy individual IDs
    n_indiv = grm_list[0].shape[0]
    n_pheno = pheno.shape[1]
    n_grm = len(grm_list)

    df_id = pd.DataFrame(
        {
            "FID": np.arange(n_indiv).astype(str),
            "IID": np.arange(n_indiv).astype(str),
        }
    )

    rls_list = []

    # create temporary directory
    with tempfile.TemporaryDirectory() as tmp_dir:
        print(tmp_dir)
        with cd(tmp_dir):
            f = open("mgrm.txt", "w")
            # the last grm corresponds to identity matrix, which will be coped
            # internally by GCTA, so we don't need to write it
            for i_grm, grm in enumerate(grm_list[:-1]):
                name = f"grm_{i_grm}"
                # NOTE: the degree of freedom for GCTA implementation is not
                # properly set up. Should not be a big problem for the current.

                # NOTE: n_snps will not be used by GCTA in this HE regression.
                # so we just set it to some arbitrary value `1`.
                admix.io.write_gcta_grm(
                    file_prefix=name,
                    grm=grm.compute(),
                    df_id=df_id,
                    n_snps=np.array([1] * n_indiv),
                )
                f.write(f"{name}\n")
            f.close()
            df_pheno = df_id.copy()
            df_covar = df_id.copy()
            df_covar[[f"COV{cov_i}" for cov_i in range(cov.shape[1])]] = cov
            df_covar.to_csv("covar.txt", sep="\t", index=False, header=False)

            # write covarites
            for i_pheno in range(n_pheno):
                df_pheno["PHENO"] = pheno[:, i_pheno]
                df_pheno.to_csv("pheno.txt", sep="\t", index=False, header=False)

                # run GCTA
                admix.tools.gcta(
                    "--reml --mgrm mgrm.txt --pheno pheno.txt --qcovar covar.txt --reml-est-fix --reml-no-constrain --out result"
                )

                df_rls = pd.read_csv(
                    "result.hsq",
                    delim_whitespace=True,
                    nrows=n_grm,
                )
                rls_list.append(df_rls)

    return rls_list


from . import get_dependency
import numpy as np
import tempfile
import urllib.request
import subprocess
from os.path import join
import pandas as pd


def run(df_chrom_pos: pd.DataFrame, chain: str, verbose: bool = False):

    """Lift over between genome assembly
    Download appropriate chain file from
    hg38->hg19 http://hgdownload.cse.ucsc.edu/goldenpath/hg38/liftOver/hg38ToHg19.over.chain.gz
    hg19->hg38 http://hgdownload.cse.ucsc.edu/goldenpath/hg19/liftOver/hg19ToHg38.over.chain.gz
    Download executable from http://hgdownload.soe.ucsc.edu/admin/exe/
    (1) chmod +x liftOver
    (2) include to the path, so the program can call liftOver

    Parameters
    ----------
    df_chrom_pos: pd.DataFrame
        where the 1st column are the chromosomes, without `chr`
        and the 2nd column are the positions
    chain : str
        chain file

    Returns
    -------
    index of df_chrom_pos and the corresponding positions
    SNP positions after the liftOver, unmapped SNPs are returned as -1
    """

    df_chrom_pos = df_chrom_pos.copy()
    df_chrom_pos.columns = ["CHROM", "POS"]
    bin_path = get_dependency("liftOver")

    assert chain in [
        "hg38->hg19",
        "hg19->hg38",
    ], "Currently only these two chain files are supported"
    url_dict = {
        "hg38->hg19": "http://hgdownload.cse.ucsc.edu/goldenpath/hg38/liftOver/hg38ToHg19.over.chain.gz",
        "hg19->hg38": "http://hgdownload.cse.ucsc.edu/goldenpath/hg19/liftOver/hg19ToHg38.over.chain.gz",
    }
    df_chrom_pos = pd.DataFrame(
        {
            "CHROM": "chr" + df_chrom_pos["CHROM"].astype(str),
            "POS0": df_chrom_pos["POS"] - 1,
            "POS": df_chrom_pos["POS"],
            "ID": df_chrom_pos.index.values.astype(str),
        }
    )

    tmp_dir = tempfile.TemporaryDirectory()
    old_file = join(tmp_dir.name, "old.bed")
    chain_file = join(tmp_dir.name, "chain.gz")
    new_file = join(tmp_dir.name, "new.bed")
    unmapped_file = join(tmp_dir.name, "unmapped.txt")

    urllib.request.urlretrieve(url_dict[chain], chain_file)
    df_chrom_pos.to_csv(old_file, sep="\t", index=False, header=False)
    cmd = f"{bin_path} {old_file} {chain_file} {new_file} {unmapped_file}"
    subprocess.check_call(cmd, shell=True)

    # read the mapped data frame, SNPs with ambiguous mapping are dropped
    df_new = pd.read_csv(
        new_file, sep="\t", header=None, names=["CHROM", "POS0", "POS", "ID"]
    ).drop_duplicates(subset=["ID"], keep=False)
    # count number of lines starting with a #
    n_unmapped = sum(1 for line in open(unmapped_file) if line.startswith("#"))
    print(f"{n_unmapped} un-mapped variants")

    tmp_dir.cleanup()

    # ambiguous mapping or unmapped SNPs are -1, return
    df_ret = (
        df_new.set_index("ID")
        .reindex(df_chrom_pos.index.values)[["POS"]]
        .fillna(-1)
        .astype(int)
    )

    df_ret.index.name = df_chrom_pos.index.name
    return df_ret


"""
admix.data is for all sorts of data manipulation
including genotype matrix, local ancestry matrix

These functions should not depend on admix.Dataset, but rather can be used
on their own alone.
"""

from ._geno import allele_per_anc, af_per_anc, geno_mult_mat
from ._geno import calc_pgs, calc_partial_pgs
from ._geno import calc_snp_prior_var, impute_with_mean
from ._geno import grm, admix_ld, admix_grm, admix_grm_equal_var

from ._stats import (
    quantile_normalize,
    pval2chisq,
    lambda_gc,
    zsc2pval,
    hdi,
    deming_regression,
    meta_analysis,
)
from ._lanc import (
    Lanc,
    concat_lancs,
    lanc_impute_single_chrom,
    haplo2diplo,
    clean_lanc,
    read_bp_lanc,
)
from ._utils import (
    index_over_chunks,
    make_dataset,
    match_prs_weights,
    distance_to_line,
    distance_to_refpop,
)
from ._misc import convert_dummy


"""Common utility functions for `admix`. Used by more than 2 modules"""
import admix
import os
from contextlib import contextmanager
import os
from typing import List
import glob
import hashlib


def str2int(s: str) -> int:
    """Convert string to int

    Parameters
    ----------
    s : str
        string to convert

    Returns
    -------
    int
        string and hashed
    """
    return int(hashlib.sha256(s.encode("utf-8")).hexdigest(), 16) % (2 ** 32)


@contextmanager
def cd(newdir):
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)


def get_cache_dir() -> str:
    """Get the cache directory for admix-kit

    Returns
    -------
    [type]
        [description]
    """
    cache_dir = os.path.join(os.path.dirname(admix.__file__), "../.admix_cache")
    os.makedirs(cache_dir, exist_ok=True)
    return cache_dir



import admix
import pandas as pd
from typing import Union, List
from admix import logger
from ._utils import log_params


def assoc(
    pfile: str,
    pheno: str,
    out: str,
    method: Union[str, List[str]] = "ATT",
    family: str = "quant",
    quantile_normalize: bool = False,
    snp_list: str = None,
    fast: bool = True,
):
    """
    Perform association testing.

    Parameters
    ----------
    pfile : str
        Prefix to the PLINK2 file (.pgen should not be added). When using a method requiring
        local ancestry, a matching :code:`<pfile>.lanc` file should also exist.
    pheno : str
        Path to the phenotype file. The text file should be space delimited with header
        and one individual per row. 1st column: individual ID. 2nd column: phenotype
        values. 3rd - nth columns: covariates. NaN should be encoded as "NA" and these
        individuals will be removed in the analysis.
        Binary phenotype should be encoded as 0 and 1, and
        :code:`--family binary` should be used.  All
        columns will be used for the analysis. NaN should be encoded as "NA" and NaN
        will be imputed with the mean of each covariate. Categorical covariates will be
        converted to one hot encodings internally.
    out : str
        Path the output file. :code:`<out>.<method>.assoc` will be created.
    method : Union[str, List[str]]
        Method to use for association analysis (default ATT). Other methods include:
        TRACTOR, ADM, SNP1, HET
    family : str
        Family to use for association analysis (default quant). One of :code:`quant` or
        :code:`binary`.
    quantile_normalize : bool
        Whether to quantile normalize the phenotype and every covariate. When
        :code:`--family binary` is used, quantile normalization will only be applied
        to covariates.
    snp_list : str
        Path to the SNP list file. Each line should be a SNP ID. Only SNPs in the
        list will be used for the analysis.
    fast : bool
        Whether to use fast mode (default True).
    """

    log_params("assoc", locals())
    assert family in ["quant", "binary"], "family must be either quant or binary"

    # TODO: infer block size using memory use in dask-pgen read
    dset = admix.io.read_dataset(pfile)
    admix.logger.info(f"{dset.n_snp} SNPs and {dset.n_indiv} individuals are loaded")
    df_pheno = pd.read_csv(pheno, delim_whitespace=True, index_col=0, low_memory=False)
    df_pheno.index = df_pheno.index.astype(str)
    pheno_col = df_pheno.columns[0]
    covar_cols = df_pheno.columns[1:]
    if len(covar_cols) == 0:
        covar_cols = None

    admix.logger.info(f"trait column: {pheno_col}")
    admix.logger.info(f"covariate columns: {covar_cols}")
    dset.append_indiv_info(df_pheno, force_update=True)

    # retain only individuals with non-missing phenotype,
    # or with non-completely missing covariate
    indiv_mask = ~dset.indiv[pheno_col].isna().values
    if covar_cols is not None:
        covar_mask = ~(dset.indiv[covar_cols].isna().values.all(axis=1))
        indiv_mask &= covar_mask
    dset = dset[:, indiv_mask]
    admix.logger.info(
        f"{dset.n_indiv} individuals left "
        "after filtering for missing phenotype, or completely missing covariate"
    )

    # filter for SNPs
    if snp_list is not None:
        with open(snp_list, "r") as f:
            filter_snp_list = [line.strip() for line in f]
        n_filter_snp = len(filter_snp_list)
        filter_snp_list = dset.snp.index[dset.snp.index.isin(filter_snp_list)]
        if len(filter_snp_list) < n_filter_snp:
            admix.logger.warning(
                f"{n_filter_snp - len(filter_snp_list)} SNPs in {snp_list} are not in the dataset"
            )
        dset = dset[filter_snp_list]

    admix.logger.info(
        f"{dset.n_snp} SNPs and {dset.n_indiv} individuals in the analysis"
    )

    if isinstance(method, str):
        method = [method]

    # extract pheno_values and covar_values
    pheno_values = dset.indiv[pheno_col].values
    if covar_cols is not None:
        covar_values = admix.data.convert_dummy(dset.indiv[covar_cols]).values
    else:
        covar_values = None

    if quantile_normalize:
        if family == "quant":
            pheno_values = admix.data.quantile_normalize(pheno_values)
        if covar_values is not None:
            for i in range(covar_values.shape[1]):
                covar_values[:, i] = admix.data.quantile_normalize(covar_values[:, i])

    dict_rls = {}
    for m in method:
        admix.logger.info(f"Performing association analysis with method {m}")
        dict_rls[m] = admix.assoc.marginal(
            dset=dset,
            pheno=pheno_values,
            cov=covar_values,
            method=m,
            family="logistic" if family == "binary" else "linear",
            fast=fast,
        )
    for m in dict_rls:
        dict_rls[m].to_csv(
            f"{out}.{m}.assoc", sep="\t", float_format="%.6g", na_rep="NA"
        )
        logger.info(f"Output written to {out}.{m}.assoc")



import numpy as np
import pandas as pd
from tqdm import tqdm
import dask.array as da
import admix
import dask
from typing import Union, Tuple, List
import dapgen


def calc_snp_prior_var(df_snp_info, her_model):
    """
    Calculate the SNP prior variance from SNP information
    """
    assert her_model in ["uniform", "gcta", "ldak", "mafukb"]
    if her_model == "uniform":
        return np.ones(len(df_snp_info))
    elif her_model == "gcta":
        freq = df_snp_info["FREQ"].values
        assert np.all(freq > 0), "frequencies should be larger than zero"
        return np.float_power(freq * (1 - freq), -1)
    elif her_model == "mafukb":
        # MAF-dependent genetic architecture, \alpha = -0.38 estimated from meta-analysis in UKB traits
        freq = df_snp_info["FREQ"].values
        assert np.all(freq > 0), "frequencies should be larger than zero"
        return np.float_power(freq * (1 - freq), -0.38)
    elif her_model == "ldak":
        freq, weight = df_snp_info["FREQ"].values, df_snp_info["LDAK_WEIGHT"].values
        return np.float_power(freq * (1 - freq), -0.25) * weight
    else:
        raise NotImplementedError


def impute_with_mean(mat, inplace=False, axis=1):
    """impute the each entry using the mean of the input matrix np.mean(mat, axis=axis)
    axis = 1 corresponds to row-wise imputation
    axis = 0 corresponds to column-wise imputation

    Parameters
    ----------
    mat : np.ndarray
        input matrix. For reminder, the genotype matrix is with shape (n_snp, n_indiv)
    inplace : bool
        whether to return a new dataset or modify the input dataset
    axis : int
        axis to impute along

    Returns
    -------
    if inplace:
        mat : np.ndarray
            (n_snp, n_indiv) matrix
    else:
        None
    """
    assert axis in [0, 1], "axis should be 0 or 1"
    if not inplace:
        mat = mat.copy()

    # impute the missing genotypes with the mean of each row
    mean = np.nanmean(mat, axis=axis)
    nanidx = np.where(np.isnan(mat))

    # index the mean using the nanidx[1 - axis]
    # axis = 1, row-wise imputation, index the mean using the nanidx[0]
    # axis = 0, columnw-ise imputation, index the mean using the nanidx[1]
    mat[nanidx] = mean[nanidx[1 - axis]]

    if not inplace:
        return mat
    else:
        return None


def geno_mult_mat(
    geno: da.Array,
    mat: np.ndarray,
    impute_geno: bool = True,
    mat_dim: str = "snp",
    return_snp_var: bool = False,
) -> np.ndarray:
    """Multiply genotype matrix with another matrix

    Chunk of genotype matrix will be read sequentially along the SNP dimension,
    and multiplied with the `mat`.

    Without transpose, result will be (n_snp, n_rep)
    With transpose, result will be (n_indiv, n_rep)

    Missing values in geno will be imputed with the mean of the genotype matrix.

    Parameters
    ----------
    geno : da.Array
        Genotype matrix with shape (n_snp, n_indiv)
        geno.chunk contains the chunk of genotype matrix to be multiplied
    mat : np.ndarray
        Matrix to be multiplied with the genotype matrix. If the passed variable
        is a vector, it will be transformed to be a 1-column matrix.
    impute_geno : bool
        Whether to impute missing values with the mean of the genotype matrix
    mat_dim : str
        First dimension of the `mat`, either "snp" or "indiv"
        Whether to transpose the genotype matrix and calulate geno.T @ mat
    return_snp_var : bool
        Whether to return the variance of each SNP, useful in simple linear
        regression

    Returns
    -------
    np.ndarray
        Result of the multiplication
    """
    assert mat_dim in ["snp", "indiv"], "mat_dim should be `snp` or `indiv`"

    if mat.ndim == 1:
        mat = mat[:, np.newaxis]
    # chunks over SNPs
    chunks = geno.chunks[0]
    indices = np.insert(np.cumsum(chunks), 0, 0)
    n_snp, n_indiv = geno.shape
    n_rep = mat.shape[1]

    snp_var = np.zeros(n_snp)
    if mat_dim == "indiv":
        # geno: (n_snp, n_indiv)
        # mat: (n_indiv, n_rep)
        assert (
            mat.shape[0] == n_indiv
        ), "when mat_dim is 'indiv', matrix should be of shape (n_indiv, n_rep)"
        ret = np.zeros((n_snp, n_rep))
        for i in tqdm(range(len(indices) - 1), desc="admix.data.geno_mult_mat"):
            start, stop = indices[i], indices[i + 1]
            geno_chunk = geno[start:stop, :].compute()
            # impute missing genotype
            if impute_geno:
                impute_with_mean(geno_chunk, inplace=True)
            ret[start:stop, :] = np.dot(geno_chunk, mat)

            if return_snp_var:
                snp_var[start:stop] = np.var(geno_chunk, axis=0)
    elif mat_dim == "snp":
        # geno: (n_indiv, n_snp)
        # mat: (n_snp, n_rep)
        assert (
            mat.shape[0] == n_snp
        ), "when mat_dim is 'snp', matrix should be of shape (n_snp, n_rep)"
        ret = np.zeros((n_indiv, n_rep))
        for i in tqdm(range(len(indices) - 1), desc="admix.data.geno_mult_mat"):
            start, stop = indices[i], indices[i + 1]
            geno_chunk = geno[start:stop, :].compute()
            # impute missing genotype
            if impute_geno:
                impute_with_mean(geno_chunk, inplace=True)
            ret += np.dot(geno_chunk.T, mat[start:stop, :])

            if return_snp_var:
                snp_var[start:stop] = np.var(geno_chunk, axis=0)
    else:
        raise ValueError("mat_dim should be `snp` or `indiv`")
    if return_snp_var:
        return ret, snp_var
    else:
        return ret


def grm(geno: da.Array, subpopu: np.ndarray = None, std_method: str = "std"):
    """Calculate the GRM matrix
    This function is to serve as an alternative of GCTA --make-grm

    Parameters
    ----------
    geno: admix.Dataset
        genotype (n_snp, n_indiv) matrix
    subpopu : np.ndarray
        subpopulation labels, with shape (n_indiv,). The allele frequencies and
        normalization are performed separately within each subpopulation.
    std_method : str
        Method to standardize the GRM. Currently supported:
        "std" (standardize to have mean 0 and variance 1),
        "allele" (standardize to have mean 0 but no scaling)

    Returns
    -------
    np.ndarray
        GRM matrix (n_indiv, n_indiv)
    """

    def normalize_geno(g):
        """Normalize the genotype matrix"""
        # impute missing genotypes
        g = impute_with_mean(g, inplace=False, axis=1)
        # normalize
        if std_method == "std":
            g = (g - np.mean(g, axis=1)[:, None]) / np.std(g, axis=1)[:, None]
        elif std_method == "allele":
            g = g - np.mean(g, axis=1)[:, None]
        else:
            raise ValueError("std_method should be either `std` or `allele`")
        return g

    assert std_method in ["std", "allele"], "std_method should be `std` or `allele`"
    n_snp = geno.shape[0]
    n_indiv = geno.shape[1]

    if subpopu is not None:
        assert (
            n_indiv == subpopu.shape[0]
        ), "subpopu should have the same length as the number of individuals"
        unique_subpopu = np.unique(subpopu)
        admix.logger.info(
            f"{len(unique_subpopu)} subpopulations found: {unique_subpopu}"
        )

    admix.logger.info(
        f"Calculating GRM matrix with {n_snp} SNPs and {n_indiv} individuals"
    )
    mat = 0
    snp_chunks = geno.chunks[0]
    indices = np.insert(np.cumsum(snp_chunks), 0, 0)
    for i in tqdm(range(len(indices) - 1), desc="admix.data.grm"):
        start, stop = indices[i], indices[i + 1]
        geno_chunk = geno[start:stop, :].compute()
        if subpopu is not None:
            for popu in np.unique(subpopu):
                geno_chunk[:, subpopu == popu] = normalize_geno(
                    geno_chunk[:, subpopu == popu]
                )
        else:
            geno_chunk = normalize_geno(geno_chunk)

        mat += np.dot(geno_chunk.T, geno_chunk) / n_snp
    return mat


def admix_grm(
    geno: da.Array, lanc: da.Array, n_anc: int = 2, snp_prior_var: np.ndarray = None
):
    """Calculate ancestry specific GRM matrix

    Parameters
    ----------
    geno : da.Array
        Genotype matrix with shape (n_snp, n_indiv, 2)
    lanc : np.ndarray
        Local ancestry matrix with shape (n_snp, n_indiv, 2)
    n_anc : int
        Number of ancestral populations
    snp_prior_var : np.ndarray
        Prior variance of each SNP, shape (n_snp,)

    Returns
    -------
    G1: np.ndarray
        ancestry specific GRM matrix for the 1st ancestry
    G2: np.ndarray
        ancestry specific GRM matrix for the 2nd ancestry
    G12: np.ndarray
        ancestry specific GRM matrix for cross term of the 1st and 2nd ancestry
    """

    assert n_anc == 2, "only two-way admixture is implemented"
    assert np.all(geno.shape == lanc.shape)

    apa = admix.data.allele_per_anc(geno, lanc, n_anc=n_anc)
    n_snp, n_indiv = apa.shape[0:2]

    if snp_prior_var is None:
        snp_prior_var = np.ones(n_snp)
    snp_prior_var_sum = snp_prior_var.sum()
    G1 = np.zeros([n_indiv, n_indiv])
    G2 = np.zeros([n_indiv, n_indiv])
    G12 = np.zeros([n_indiv, n_indiv])

    snp_chunks = apa.chunks[0]
    indices = np.insert(np.cumsum(snp_chunks), 0, 0)

    for i in tqdm(range(len(indices) - 1), desc="admix.data.admix_grm"):
        start, stop = indices[i], indices[i + 1]
        apa_chunk = apa[start:stop, :, :].compute()

        # multiply by the prior variance on each SNP
        apa_chunk *= np.sqrt(snp_prior_var[start:stop])[:, None, None]
        a1_chunk, a2_chunk = apa_chunk[:, :, 0], apa_chunk[:, :, 1]

        G1 += np.dot(a1_chunk.T, a1_chunk) / snp_prior_var_sum
        G2 += np.dot(a2_chunk.T, a2_chunk) / snp_prior_var_sum
        G12 += np.dot(a1_chunk.T, a2_chunk) / snp_prior_var_sum

    return G1, G2, G12


def admix_grm_equal_var(
    geno: da.Array, lanc: da.Array, n_anc: int, snp_prior_var: np.ndarray = None
):
    """Calculate ancestry specific GRM matrix K1, K2 (assuming equal variances for ancestries)

    Parameters
    ----------
    geno : da.Array
        Genotype matrix with shape (n_snp, n_indiv, 2)
    lanc : np.ndarray
        Local ancestry matrix with shape (n_snp, n_indiv, 2)
    n_anc : int
        Number of ancestral populations
    snp_prior_var : np.ndarray
        Prior variance of each SNP, shape (n_snp,)

    Returns
    -------
    K1: np.ndarray
        sum of diagonal terms
    K2: np.ndarray
        off-diagonal terms
    """
    assert np.all(geno.shape == lanc.shape)

    apa = admix.data.allele_per_anc(geno, lanc, n_anc=n_anc)
    n_snp, n_indiv = apa.shape[0:2]

    if snp_prior_var is None:
        snp_prior_var = np.ones(n_snp)
    snp_prior_var_sum = snp_prior_var.sum()

    K1 = np.zeros([n_indiv, n_indiv])
    K2 = np.zeros([n_indiv, n_indiv])

    snp_chunks = apa.chunks[0]
    indices = np.insert(np.cumsum(snp_chunks), 0, 0)

    for i in tqdm(range(len(indices) - 1), desc="admix.data.admix_grm_equal_var"):
        start, stop = indices[i], indices[i + 1]
        apa_chunk = apa[start:stop, :, :].compute()

        # multiply by the prior variance on each SNP
        apa_chunk *= np.sqrt(snp_prior_var[start:stop])[:, None, None]

        # diagonal terms
        for i_anc in range(n_anc):
            a_chunk = apa_chunk[:, :, i_anc]
            K1 += np.dot(a_chunk.T, a_chunk) / snp_prior_var_sum

        # off-diagonal terms
        for i_anc in range(n_anc):
            for j_anc in range(i_anc + 1, n_anc):
                a1_chunk, a2_chunk = apa_chunk[:, :, i_anc], apa_chunk[:, :, j_anc]
                K2 += np.dot(a1_chunk.T, a2_chunk) / snp_prior_var_sum

    K2 = K2 + K2.T
    return K1, K2


def admix_ld(dset: admix.Dataset, cov: np.ndarray = None):
    """Calculate ancestry specific LD matrices

    Parameters
    ----------
    dset: admix.Dataset
        dataset containing geno, lanc
    cov : Optional[np.ndarray]
        (n_indiv, n_cov) covariates of the genotypes, an all `1` intercept covariate will always be added
        so there is no need to add the intercept in covariates.
    Returns
    -------
    K1: np.ndarray
        ancestry specific LD matrix for the 1st ancestry
    K2: np.ndarray
        ancestry specific LD matrix for the 2nd ancestry
    K12: np.ndarray
        ancestry specific LD matrix for cross term of the 1st and 2nd ancestry
    """
    assert dset.n_anc == 2, "admix_ld only works for 2 ancestries for now"
    apa = dset.allele_per_anc()

    n_snp, n_indiv = apa.shape[0:2]

    a1, a2 = apa[:, :, 0], apa[:, :, 1]
    if cov is None:
        cov = np.ones((n_indiv, 1))
    else:
        cov = np.hstack([np.ones((n_indiv, 1)), cov])
    # projection = I - X * (X'X)^-1 * X'
    cov_proj_mat = np.eye(n_indiv) - np.linalg.multi_dot(
        [cov, np.linalg.inv(np.dot(cov.T, cov)), cov.T]
    )
    a1 = np.dot(a1, cov_proj_mat)
    a2 = np.dot(a2, cov_proj_mat)
    # center with row mean
    # a1 -= a1.mean(axis=1, keepdims=True)
    # a2 -= a2.mean(axis=1, keepdims=True)
    ld1 = np.dot(a1, a1.T) / n_indiv
    ld2 = np.dot(a2, a2.T) / n_indiv
    ld12 = np.dot(a1, a2.T) / n_indiv
    ld1, ld2, ld12 = dask.compute(ld1, ld2, ld12)
    return {"11": ld1, "22": ld2, "12": ld12}


def af_per_anc(
    geno, lanc, n_anc=2, return_nhaplo=False
) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """
    Calculate allele frequency per ancestry

    If at one particular SNP locus, no SNP from one particular ancestry can be found
    the corresponding entries will be filled with np.NaN.

    Parameters
    ----------
    geno: np.ndarray
        genotype matrix
    lanc: np.ndarray
        local ancestry matrix
    n_anc: int
        number of ancestries
    return_nhaplo: bool
        whether to return the number of haplotypes per ancestry

    Returns
    -------
    np.ndarray
        (n_snp, n_anc) length list of allele frequencies.
    """
    assert np.all(geno.shape == lanc.shape)
    n_snp = geno.shape[0]
    af = np.zeros((n_snp, n_anc))
    lanc_nhaplo = np.zeros((n_snp, n_anc))
    snp_chunks = geno.chunks[0]
    indices = np.insert(np.cumsum(snp_chunks), 0, 0)

    for i in tqdm(range(len(indices) - 1), desc="admix.data.af_per_anc"):
        start, stop = indices[i], indices[i + 1]
        geno_chunk = geno[start:stop, :, :].compute()
        lanc_chunk = lanc[start:stop, :, :].compute()

        for anc_i in range(n_anc):
            lanc_mask = lanc_chunk == anc_i
            lanc_nhaplo[start:stop, anc_i] = np.sum(lanc_mask, axis=(1, 2))
            # mask SNPs with local ancestry not `i_anc`
            af[start:stop, anc_i] = (
                np.ma.masked_where(np.logical_not(lanc_mask), geno_chunk)
                .sum(axis=(1, 2))
                .data
            ) / lanc_nhaplo[start:stop, anc_i]

    if return_nhaplo:
        return af, lanc_nhaplo
    else:
        return af


def allele_per_anc(
    geno: da.Array,
    lanc: da.Array,
    n_anc: int,
    center=False,
):
    """Get allele count per ancestry

    Parameters
    ----------
    geno: da.Array
        genotype data
    lanc: da.Array
        local ancestry data
    n_anc: int
        number of ancestries

    Returns
    -------
    Return allele counts per ancestries
    """
    assert center is False, "center=True should not be used"
    assert np.all(geno.shape == lanc.shape), "shape of `hap` and `lanc` are not equal"
    assert geno.ndim == 3, "`hap` and `lanc` should have three dimension"
    n_snp, n_indiv, n_haplo = geno.shape
    assert n_haplo == 2, "`n_haplo` should equal to 2, check your data"

    assert isinstance(geno, da.Array) & isinstance(
        lanc, da.Array
    ), "`geno` and `lanc` should be dask array"

    # make sure the chunk size along the ploidy axis to be 2
    geno = geno.rechunk({2: 2})
    lanc = lanc.rechunk({2: 2})

    assert (
        geno.chunks == lanc.chunks
    ), "`geno` and `lanc` should have the same chunk size"

    assert len(geno.chunks[1]) == 1, (
        "geno / lanc should not be chunked across the second dimension"
        "(individual dimension)"
    )

    def helper(geno_chunk, lanc_chunk, n_anc):
        n_snp, n_indiv, n_haplo = geno_chunk.shape
        apa = np.zeros((n_snp, n_indiv, n_anc), dtype=np.float64)
        for i_haplo in range(n_haplo):
            haplo_hap = geno_chunk[:, :, i_haplo]
            haplo_lanc = lanc_chunk[:, :, i_haplo]
            for i_anc in range(n_anc):
                apa[:, :, i_anc][haplo_lanc == i_anc] += haplo_hap[haplo_lanc == i_anc]
        return apa

    # the resulting chunk sizes will be the same as the input for snp, indiv
    # while the third dimension will be (n_anc, )
    output_chunks = (geno.chunks[0], geno.chunks[1], (n_anc,))
    res = da.map_blocks(
        lambda geno_chunk, lanc_chunk: helper(
            geno_chunk=geno_chunk, lanc_chunk=lanc_chunk, n_anc=n_anc
        ),
        geno,
        lanc,
        dtype=np.float64,
        chunks=output_chunks,
    )

    return res


def calc_pgs(dset: admix.Dataset, df_weights: pd.DataFrame, method: str):
    """Calculate PGS for each individual

    Parameters
    ----------
    dset: admix.Dataset
        dataset object
    df_weights: pd.DataFrame
        weights for each individual
    method: str
        method to calculate PGS. Options are:
        - "total": vanilla PGS
        - "partial": partial PGS, calculate partial PGS for each local ancestry

    Returns
    -------
    np.ndarray
        PGS for each individual
        - method = "total": (n_indiv, )
        - method = "partial": (n_indiv, n_anc)
    """
    assert method in [
        "total",
        "partial",
    ], "method should be either 'total' or 'partial'"
    assert np.all(
        dset.snp.index == df_weights.index
    ), "`dset` and `df_weights` should have exactly the same index"

    assert len(df_weights.columns) == 1, "`df_weights` should have only one column"

    if method == "total":
        pgs = admix.data.geno_mult_mat(
            dset.geno.sum(axis=2), df_weights.values
        ).flatten()
    elif method == "partial":
        n_anc = dset.n_anc
        pgs = np.zeros((dset.n_indiv, n_anc))
        apa = dset.allele_per_anc()
        for i_anc in range(n_anc):
            pgs[:, i_anc] = admix.data.geno_mult_mat(
                apa[:, :, i_anc], df_weights.values
            ).flatten()

    else:
        raise ValueError("method should be either 'total' or 'partial'")

    return pgs


def calc_partial_pgs(
    dset: admix.Dataset,
    df_weights: pd.DataFrame,
    dset_ref: admix.Dataset = None,
    ref_pop_indiv: List[List[str]] = None,
    weight_col="WEIGHT",
) -> pd.DataFrame:
    """Calculate PGS for each individual

    Parameters
    ----------
    dset: admix.Dataset
        dataset object
    df_weights: pd.DataFrame
        weights for each individual
    dset_ref: admix.Dataset
        reference dataset object, use `dapgen.align_snp` to align the SNPs between
        `dset` and `dset_ref`. `CHROM` and `POS` must match, with potential flips of
        `REF` and `ALT` allele coding.
    ref_pop: List[List[str]]
        list of reference individual ID in `dset_ref`

    Returns
    -------
    pd.DataFrame
        PGS for each individual
        - (n_indiv, n_anc)
    """
    assert (dset_ref is None) == (
        ref_pop_indiv is None
    ), "both `dset_ref` and `ref_pop_indiv` should be None or not None"
    CALC_REF = dset_ref is not None
    CHECK_COLS = ["CHROM", "POS", "REF", "ALT"]
    ## check input
    idx1, idx2, sample_wgt_flip = dapgen.align_snp(
        df1=dset.snp[CHECK_COLS], df2=df_weights[CHECK_COLS]
    )

    assert np.all(idx1 == dset.snp.index) & np.all(
        idx2 == df_weights.index
    ), "`dset` and `df_weights` should align, with potential allele flip"

    if CALC_REF:
        idx1, idx2, ref_wgt_flip = dapgen.align_snp(
            df1=dset.snp[CHECK_COLS], df2=dset_ref.snp[CHECK_COLS]
        )
        assert np.all(idx1 == dset.snp.index) & np.all(
            idx2 == dset_ref.snp.index
        ), "`dset` and `dset_ref` should align, with potential allele flip"

    weights = df_weights[weight_col].values
    sample_weights = weights * sample_wgt_flip

    if CALC_REF:
        ref_weights = weights * ref_wgt_flip * sample_wgt_flip

        assert (
            len(ref_pop_indiv) == dset.n_anc
        ), "`len(ref_pops)` should match with `dset.n_anc`"

    ## scoring
    dset_geno, dset_lanc = dset.geno.compute(), dset.lanc.compute()
    sample_pgs = np.zeros((dset.n_indiv, dset.n_anc))
    if CALC_REF:
        ref_geno_list = [dset_ref[:, pop].geno.compute() for pop in ref_pop_indiv]
        ref_pgs = [[] for pop in ref_pop_indiv]
    # iterate over each individuals
    for indiv_i in tqdm(range(dset.n_indiv), desc="admix.data.calc_partial_pgs"):
        indiv_ref_pgs = [0, 0]
        # pgs for sample individuals
        for haplo_i in range(2):
            geno = dset_geno[:, indiv_i, haplo_i]
            lanc = dset_lanc[:, indiv_i, haplo_i]
            for lanc_i in range(dset.n_anc):
                # sample
                sample_pgs[indiv_i, lanc_i] += np.dot(
                    geno[lanc == lanc_i], sample_weights[lanc == lanc_i]
                )

                # pgs for reference individuals
                if CALC_REF:
                    ref_geno = ref_geno_list[lanc_i][lanc == lanc_i, :, :]
                    if ref_geno.shape[0] > 0:
                        ref_geno = ref_geno.reshape(ref_geno.shape[0], -1)
                        s = np.dot(ref_weights[lanc == lanc_i], ref_geno)
                    else:
                        s = np.zeros(ref_geno.shape[1] * 2)
                    indiv_ref_pgs[lanc_i] += s
        if CALC_REF:
            for lanc_i in range(dset.n_anc):
                ref_pgs[lanc_i].append(indiv_ref_pgs[lanc_i])

    # format ref_pgs: for each ancestry, we have n_indiv x (n_ref_indiv x 2)
    # each reference has 2 haplotypes
    if CALC_REF:
        ref_pgs = [
            pd.DataFrame(
                data=np.vstack(ref_pgs[i]),
                index=dset.indiv.index,
                columns=np.concatenate(
                    [[str(i) + "_1", str(i) + "_2"] for i in ref_pop_indiv[i]]
                ),
            )
            for i in range(dset.n_anc)
        ]
    sample_pgs = pd.DataFrame(
        data=sample_pgs,
        index=dset.indiv.index,
        columns=[f"ANC{i}" for i in range(1, dset.n_anc + 1)],
    )
    if CALC_REF:
        return sample_pgs, ref_pgs
    else:
        return sample_pgs


import numpy as np
from typing import List
import admix

"""
This will be the interface to call various local ancestry inference methods
"""


def lanc_single_chrom(
    sample_dset: admix.Dataset,
    ref_dsets: List[admix.Dataset],
    method="lampld",
    n_proto=6,
    window_size=300,
):
    """Local ancestry inference for data sets of single chromosome

    Parameters
    ----------
    sample_dset : admix.Dataset
        dataset of the sample to be inferred
    ref_dsets : List[admix.Dataset]
        dataset of the reference panel
    method : str, optional
        method used to perform local ancestry inference, by default "lampld"
    n_proto : int, optional
        number of prototypes used in lampld, by default 6
    window_size : int, optional
    """
    from pylampld import LampLD

    assert method in ["lampld"]
    assert len(sample_dset.snp.CHROM.unique()), "Datasets are not of single chromosome"
    assert admix.dataset.is_aligned(
        [sample_dset, *ref_dsets]
    ), "Datasets are not aligned"

    n_anc = len(ref_dsets)
    n_snp = sample_dset.n_snp

    model = LampLD(n_snp=n_snp, n_anc=n_anc, n_proto=n_proto, window_size=window_size)
    ref_list = [
        np.vstack(
            [
                dset.geno[:, :, 0].T,
                dset.geno[:, :, 1].T,
            ]
        )
        for dset in ref_dsets
    ]

    model.set_pos(sample_dset.snp.POS.values)
    model.fit(ref_list)

    est = np.dstack(
        [
            model.infer_lanc(sample_dset.geno[:, :, 0].T.compute()),
            model.infer_lanc(sample_dset.geno[:, :, 1].T.compute()),
        ]
    )
    return np.swapaxes(est.astype(np.int8), 0, 1)


def lanc(
    sample_dset: admix.Dataset,
    ref_dsets: List[admix.Dataset],
    method="lampld",
    n_proto=6,
    window_size=300,
) -> np.ndarray:
    """Local ancestry inference

    Parameters
    ----------
    sample_dset : admix.Dataset
        dataset of the sample to be inferred
    ref_dsets : List[admix.Dataset]
        dataset of the reference panel
    method : str, optional
        method used to perform local ancestry inference, by default "lampld"
    n_proto : int, optional
        number of prototypes used in lampld, by default 6
    window_size : int, optional
        window size used in local ancestry inference, by default 100

    Returns
    -------
    np.ndarray
        Estimated local ancestries
    """

    assert method in ["lampld"]
    chrom_list = sample_dset.snp.CHROM.unique()
    lanc = []
    for chrom in chrom_list:
        chrom_sample_dset = sample_dset[(sample_dset.snp.CHROM == chrom).values]
        chrom_ref_dsets = [dset[(dset.snp.CHROM == chrom).values] for dset in ref_dsets]
        chrom_lanc = lanc_single_chrom(
            chrom_sample_dset,
            chrom_ref_dsets,
            method=method,
            n_proto=n_proto,
            window_size=window_size,
        )
        lanc.append(chrom_lanc)
    return np.concatenate(lanc, axis=0)



from ._pheno import quant_pheno, quant_pheno_1pop, quant_pheno_grm
from ._pheno import binary_pheno, sample_case_control
from ._lanc import hap_lanc, calculate_mosaic_size
from ._geno import admix_geno, admix_geno_simple

# from ._genotype import simulate_hap

__all__ = [
    "quant_pheno",
    "binary_pheno",
    "sample_case_control",
    "quant_pheno_1pop",
    "quant_pheno_grm",
    "admix_geno",
]



import numpy as np
from scipy import linalg
import pandas as pd
import xarray as xr
from typing import List, Union
import admix
import tempfile
from ..utils import cd


def trace_mul(a, b):
    """
    Trace of two matrix inner product
    """
    assert np.all(a.shape == b.shape)
    return np.sum(a.flatten() * b.flatten())


def gen_cor(
    dset: admix.Dataset,
    grm: Union[str, List[str], dict],
    pheno: np.ndarray,
    cov_cols: List[str] = None,
    cov_intercept: bool = True,
    method: str = "HE",
):
    """Estimate genetic correlation given a dataset, admixture GRM, phenotypes, and covariates.

    Parameters
    ----------
    dset: admix.Dataset
        Dataset to estimate correlation from.
    grm: str, list of str or dict
        column name(s) of GRM, or a dict of {name: grm}
        Don't include the identity matrix, the indentify matrix representing environemntal
        factor will be added to the list automatically
    pheno: np.ndarray
        Phenotypes to estimate genetic correlation. If a matrix is provided, then each
        column is treated as a separate phenotype.
    cov_cols: list, optional
        List of covariate columns.
    cov_intercept: bool, optional
        Whether to include intercept in covariate matrix.
    method: str, optional
        Method to use for estimation. Valid options are:
            - "HE" (default): Haseman–Elston (HE) regression.
            - "RHE": Randomized Haseman–Elston (RHE) regression.
            - "REML": REML regression.
            - "HE-gcta": HE regression as implemented in GCTA.
            - "REML-gcta": REML as implemented in GCTA.
    """
    assert method in ["HE", "RHE", "REML"]
    n_indiv = dset.dims["indiv"]
    # get the dictionary of all the GRMs (including environmental)
    # name -> grm
    if isinstance(grm, dict):
        # obtain from the list
        grm_dict = {k: grm[k] for k in grm}
    elif isinstance(grm, list):
        grm_dict = {k: dset[k].data for k in grm}
    elif isinstance(grm, str):
        grm_dict = {grm: dset[grm].data}
    else:
        raise ValueError("`grm` must be a dictionary, list or string")

    # add the environmental GRM
    n_indiv = dset.dims["indiv"]
    grm_dict["e"] = np.eye(n_indiv)

    # build the covariate matrix
    if cov_cols is not None:
        cov_values = np.zeros((n_indiv, len(cov_cols)))
        for i_cov, cov_col in enumerate(cov_cols):
            cov_values[:, i_cov] = dset[cov_col + "@indiv"].values
        if cov_intercept:
            cov_values = np.c_[np.ones((n_indiv, 1)), cov_values]
    else:
        cov_values = None
        if cov_intercept:
            cov_values = np.ones((n_indiv, 1))

    # build projection matrix from covariate matrix
    if cov_values is None:
        cov_proj_mat = np.eye(n_indiv)
    else:
        cov_proj_mat = np.eye(n_indiv) - np.linalg.multi_dot(
            [cov_values, np.linalg.inv(np.dot(cov_values.T, cov_values)), cov_values.T]
        )

    if pheno.ndim == 1:
        pheno = pheno.reshape((-1, 1))
    assert pheno.shape[0] == n_indiv

    n_pheno = pheno.shape[1]

    pheno = np.dot(cov_proj_mat, pheno)
    quad_form_func = lambda x, A: np.dot(np.dot(x.T, A), x)

    if method == "HE":
        df_rls = []
        grm_list = [np.dot(grm_dict[name], cov_proj_mat) for name in grm_dict]
        n_grm = len(grm_list)
        HE_design = np.zeros((n_grm, n_grm))

        # fill in `HE_design`
        for i in range(n_grm):
            for j in range(n_grm):
                if i <= j:
                    HE_design[i, j] = (grm_list[i] * grm_list[j]).sum()
                    HE_design[j, i] = HE_design[i, j]

        # build response vector for each phenotype
        for i_pheno in range(n_pheno):
            HE_response = np.zeros(n_grm)
            for i in range(n_grm):
                HE_response[i] = quad_form_func(pheno[:, i_pheno], grm_list[i])
            df_rls.append(linalg.solve(HE_design, HE_response))
        df_rls = pd.DataFrame(np.vstack(df_rls), columns=[name for name in grm_dict])

    return df_rls


"""
This file is not used anywhere in current version of the code.
"""

import numpy as np
import statsmodels.api as sm
from scipy import stats


def linear_f_test1(var, cov, pheno, var_size, test_vars):
    n_indiv = var.shape[0]
    n_var = var.shape[1]
    n_cov = cov.shape[1]

    design = np.zeros((n_indiv, var_size + n_cov))
    design[:, var_size : var_size + n_cov] = cov

    n_test = int(n_var / var_size)
    pvalues = np.zeros(n_test)

    for i_test in range(n_test):
        design[:, 0:var_size] = var[:, i_test * var_size : (i_test + 1) * var_size]
        beta = np.linalg.lstsq(design, pheno, rcond=None)[0]
        sigma = np.square(pheno - np.dot(design, beta)).sum() / (
            n_indiv - var_size - n_cov
        )
        iXtX = np.linalg.inv(np.dot(design.T, design))
        f_stat = (
            beta[test_vars]
            .T.dot(np.linalg.inv(iXtX[np.ix_(test_vars, test_vars)]))
            .dot(beta[test_vars])
        ) / (len(test_vars) * sigma)
        pvalues[i_test] = stats.f.sf(f_stat, len(test_vars), n_indiv - var_size - n_cov)

    return pvalues


def linear_f_test2(var, cov, pheno, var_size, test_vars):
    n_indiv = var.shape[0]
    n_var = var.shape[1]
    n_cov = cov.shape[1]

    design = np.zeros((n_indiv, var_size + n_cov))
    design[:, var_size : var_size + n_cov] = cov

    n_test = int(n_var / var_size)
    pvalues = np.zeros(n_test)

    reduced_index = np.concatenate(
        [
            [i for i in range(var_size) if i not in test_vars],
            np.arange(var_size, var_size + n_cov),
        ]
    ).astype(int)

    for i_test in range(n_test):
        design[:, 0:var_size] = var[:, i_test * var_size : (i_test + 1) * var_size]

        beta1 = np.linalg.lstsq(design[:, reduced_index], pheno, rcond=None)[0]
        ss1 = np.square(pheno - np.dot(design[:, reduced_index], beta1)).sum()
        p1 = len(beta1)
        beta2 = np.linalg.lstsq(design, pheno, rcond=None)[0]
        ss2 = np.square(pheno - np.dot(design, beta2)).sum()
        p2 = len(beta2)

        fstat = ((ss1 - ss2) / (p2 - p1)) / (ss2 / (n_indiv - p2))
        pvalues[i_test] = stats.f.sf(fstat, p2 - p1, n_indiv - p2)

    return pvalues


def linear_lrt(var, cov, pheno, var_size, test_vars):
    n_indiv = var.shape[0]
    n_var = var.shape[1]
    n_cov = cov.shape[1]

    design = np.zeros((n_indiv, var_size + n_cov))
    design[:, var_size : var_size + n_cov] = cov

    n_test = int(n_var / var_size)
    pvalues = np.zeros(n_test)

    reduced_index = np.concatenate(
        [
            [i for i in range(var_size) if i not in test_vars],
            np.arange(var_size, var_size + n_cov),
        ]
    ).astype(int)

    for i_test in range(n_test):
        design[:, 0:var_size] = var[:, i_test * var_size : (i_test + 1) * var_size]
        llf1 = sm.OLS(pheno, design[:, reduced_index]).fit().llf
        llf2 = sm.OLS(pheno, design).fit().llf
        pvalues[i_test] = stats.chi2.sf(-2 * (llf1 - llf2), len(test_vars))
    return pvalues


def linear_f_test3(var, cov, pheno, var_size, test_vars):
    n_indiv = var.shape[0]
    n_var = var.shape[1]
    n_cov = cov.shape[1]

    design = np.zeros((n_indiv, var_size + n_cov))
    design[:, var_size : var_size + n_cov] = cov

    n_test = int(n_var / var_size)
    pvalues = np.zeros(n_test)

    f_test_r_matrix = np.zeros((len(test_vars), design.shape[1]))
    for i, v in enumerate(test_vars):
        f_test_r_matrix[i, v] = 1

    for i_test in range(n_test):
        design[:, 0:var_size] = var[:, i_test * var_size : (i_test + 1) * var_size]
        model = sm.OLS(pheno, design, missing="drop").fit()
        print(model.f_test(f_test_r_matrix))
        pvalues[i_test] = model.f_test(f_test_r_matrix).pvalue.item()
    return pvalues


import admix
import pandas as pd
import numpy as np
from typing import List
import dapgen
from ._utils import log_params


def simulate_admix_pheno(
    pfile: str,
    hsq: float,
    out_prefix: str,
    cor: float = 1.0,
    family: str = "quant",
    n_causal: int = None,
    p_causal: float = None,
    case_prevalence: float = 0.5,
    seed: int = None,
    snp_effect: str = None,
    n_sim: int = 10,
):
    """
    Simulate phenotypes using a pgen file for admixed individuals.

    Parameters
    ----------
    pfile : str
        Path to the pgen file
    hsq : float
        Heritability
    out_prefix : str
        prefix to the output file, <out>.pheno, <out>.snpeffect will be created
    cor : float
        genetic correlation across local ancestries
    family : str
        phenotype type to simulate, either "quant" or "binary"
    n_causal : int
        Number of causal variants to simulate
    p_causal : float
        Proportion of a causal variant
    case_prevalence: float
        Prevalence of cases, default 0.5
    seed : int
        Random seed
    beta : str
        Path to the beta file
    n_sim : int
        Number of simulations to perform
    """
    log_params("simulate-pheno", locals())
    assert snp_effect is None, "snp_effect is not supported yet"
    if seed is not None:
        np.random.seed(seed)

    assert (n_causal is not None) + (
        p_causal is not None
    ) == 1, "one of n_causal/p_causal must be specified"

    dset = admix.io.read_dataset(pfile)

    if p_causal is not None:
        n_causal = int(dset.n_indiv * p_causal)

    if family == "quant":
        dict_sim = admix.simulate.quant_pheno(
            dset=dset,
            hsq=hsq,
            cor=cor,
            n_causal=n_causal,
            n_sim=n_sim,
            beta=snp_effect,
        )
    elif family == "binary":
        dict_sim = admix.simulate.binary_pheno(
            dset=dset,
            hsq=hsq,
            cor=cor,
            case_prevalence=case_prevalence,
            n_causal=n_causal,
            n_sim=n_sim,
            beta=snp_effect,
        )
    else:
        raise ValueError(f"Unknown family: {family}")
    df_snp = dset.snp.copy()
    df_indiv = dset.indiv.copy()
    columns = []
    for sim_i in range(n_sim):
        columns.extend([f"SIM{sim_i}.ANC{anc_i}" for anc_i in range(dset.n_anc)])

    df_beta = pd.DataFrame(index=df_snp.index, columns=columns)
    # fill in beta
    for anc_i in range(dset.n_anc):
        df_beta.iloc[:, anc_i :: dset.n_anc] = dict_sim["beta"][:, anc_i, :]

    df_pheno = pd.DataFrame(
        dict_sim["pheno"],
        columns=[f"SIM{i}" for i in range(n_sim)],
        index=df_indiv.index,
    )

    for suffix, df in zip(["beta", "pheno"], [df_beta, df_pheno]):
        df.to_csv(f"{out_prefix}.{suffix}", sep="\t", float_format="%.6g")


def simulate_pheno(
    plink_path: str,
    hsq: float,
    out: str,
    var_e: float = None,
    family: str = "quant",
    n_causal: int = None,
    p_causal: float = None,
    case_prevalence: float = 0.5,
    seed: int = None,
    snp_effect: str = None,
    n_sim: int = 10,
):

    """Simulate phenotypes from plink files

    Parameters
    ----------
    plink_path : str
        path to plink file
    hsq : float
        simulated heritability
    var_e: float
        variance of environmental noise
    out : str
        output prefix
    family : str, optional
        distribution family of y, by default "quant"
    n_causal : int, optional
        number of causal variants, by default None
    p_causal : float, optional
        proportion of causal variants, by default None
    case_prevalence : float, optional
        prevalence of cases, by default 0.5
    seed : int, optional
        random seed, by default None
    snp_effect : str, optional
        SNP effects file, by default None
    n_sim : int, optional
        number of simulations, by default 10
    """
    log_params("simulate-pheno", locals())

    ## check input
    assert snp_effect is None, "snp_effect is not supported yet"
    assert var_e is None, "var_e is not supported yet"
    if seed is not None:
        np.random.seed(seed)

    assert (n_causal is not None) + (
        p_causal is not None
    ) == 1, "one of n_causal/p_causal must be specified"

    geno, df_snp, df_indiv = dapgen.read_plink(plink_path)
    if n_causal is None:
        n_causal = int(df_snp.shape[0] * p_causal)

    dict_simu = admix.simulate.quant_pheno_1pop(
        geno=geno, hsq=hsq, n_causal=n_causal, n_sim=n_sim
    )
    columns = [f"SIM{i}" for i in range(n_sim)]
    df_beta = pd.DataFrame(data=dict_simu["beta"], index=df_snp.index, columns=columns)
    df_pheno_g = pd.DataFrame(
        dict_simu["pheno_g"],
        columns=columns,
        index=df_indiv.index,
    )
    df_pheno = pd.DataFrame(
        dict_simu["pheno"],
        columns=columns,
        index=df_indiv.index,
    )
    admix.logger.info(f"results written to {out}.[beta|pheno_g|pheno].tsv")
    for suffix, df in zip(
        ["beta", "pheno_g", "pheno"], [df_beta, df_pheno_g, df_pheno]
    ):
        df.to_csv(f"{out}.{suffix}.tsv", sep="\t", float_format="%.6g")


import numpy as np
import pandas as pd
import statsmodels.api as sm
import admix


def partial_pgs(dset: admix.Dataset, weight: np.ndarray):
    """
    Calculate partial PGS for admix individuals

    Parameters
    ----------
    dset : admix.Dataset
        Dataset object
    weight : np.ndarray
        PGS weights

    Returns
    -------
    pgs : np.ndarray
        Partial PGS for each individual

    References
    ----------
    .. [1] Marnetto et al. (2020) Nat Commun.
    """
    pass


import numpy as np
from scipy import linalg
import xarray as xr
from typing import List, Tuple
import admix


def admix_gen_cor(
    dset: xr.Dataset,
    pheno: np.ndarray,
    cov_cols: List[str] = None,
    cov_intercept: bool = True,
):
    """Estimate genetic correlation given a dataset, phenotypes, and covariates.
    This is a very specialized function that tailed for estimating the genetic correlation
    for variants in different local ancestry backgrounds.

    See details in https://www.nature.com/articles/s41467-020-17576-9#MOESM1

    Parameters
    ----------
    dset: xr.Dataset
        Dataset to estimate correlation from.
    pheno: np.ndarray
        Phenotypes to estimate genetic correlation. If a matrix is provided, then each
        column is treated as a separate phenotype.
    cov_cols: list, optional
        List of covariate columns.
    cov_intercept: bool, optional
        Whether to include intercept in covariate matrix.
    """
    n_indiv = dset.dims["indiv"]

    # build the covariate matrix
    if cov_cols is not None:
        cov_values = np.zeros((n_indiv, len(cov_cols)))
        for i_cov, cov_col in enumerate(cov_cols):
            cov_values[:, i_cov] = dset[cov_col].values
        if cov_intercept:
            cov_values = np.c_[np.ones((n_indiv, 1)), cov_values]
    else:
        cov_values = None
        if cov_intercept:
            cov_values = np.ones((n_indiv, 1))

    # build projection matrix from covariate matrix
    if cov_values is None:
        cov_proj_mat = np.eye(n_indiv)
    else:
        cov_proj_mat = np.eye(n_indiv) - np.linalg.multi_dot(
            [cov_values, np.linalg.inv(np.dot(cov_values.T, cov_values)), cov_values.T]
        )

    if pheno.ndim == 1:
        pheno = pheno.reshape((-1, 1))
    assert pheno.shape[0] == n_indiv

    n_pheno = pheno.shape[1]

    pheno = np.dot(cov_proj_mat, pheno)
    quad_form_func = lambda x, A: np.dot(np.dot(x.T, A), x)
    n_indiv = dset.dims["indiv"]
    n_snp = dset.dims["snp"]
    # apa = admix.tools.allele_per_anc(dset, center=True, inplace=False).astype(float)
    # a1, a2 = apa[:, :, 0], apa[:, :, 1]
    # grm_list = [
    #     (np.dot(a1, a1.T) + np.dot(a2, a2.T)) / n_snp,
    #     (np.dot(a1, a2.T) + np.dot(a1, a2.T).T) / n_snp,
    #     np.eye(n_indiv),
    # ]

    grm_list = [dset["A1"].data, dset["A2"].data, np.eye(n_indiv)]
    grm_list = [np.dot(grm, cov_proj_mat) for grm in grm_list]

    # multiply cov_proj_mat
    n_grm = len(grm_list)
    design = np.zeros((n_grm, n_grm))
    for i in range(n_grm):
        for j in range(n_grm):
            if i <= j:
                design[i, j] = (grm_list[i] * grm_list[j]).sum()
                design[j, i] = design[i, j]

    rls_list: List[Tuple] = []
    for i_pheno in range(n_pheno):
        response = np.zeros(n_grm)
        for i in range(n_grm):
            response[i] = quad_form_func(pheno[:, i_pheno], grm_list[i])

        # point estimate
        var_comp = linalg.solve(
            design,
            response,
        )

        # variance-covariance matrix
        inv_design = linalg.inv(design)
        Sigma = np.zeros_like(grm_list[0])
        for i in range(n_grm):
            Sigma += var_comp[i] * grm_list[i]
        Sigma_grm_list = [np.dot(Sigma, grm) for grm in grm_list]

        var_response = np.zeros((n_grm, n_grm))
        for i in range(n_grm):
            for j in range(n_grm):
                if i <= j:
                    var_response[i, j] = (
                        2 * (Sigma_grm_list[i] * Sigma_grm_list[j]).sum()
                    )
                    var_response[j, i] = var_response[i, j]
        var_comp_var = np.linalg.multi_dot([inv_design, var_response, inv_design])
        rls_list.append((var_comp, var_comp_var))
    return rls_list


"""
Load existing data sets
"""
from typing import List
from os.path import dirname, join
import numpy as np
import os
import xarray as xr
from tqdm import tqdm
import pandas as pd
import warnings
from ._dataset import Dataset
import admix


def get_test_data_dir() -> str:
    """
    Get toy dataset directory

    Returns
    -------
    str
        Toy dataset directory
    """
    test_data_path = join(dirname(__file__), "../../tests/test-data")
    return test_data_path


def download_simulated_example_data(dir=None) -> Dataset:
    """
    Load example data set

    Returns
    -------
    Dataset
    """
    import urllib
    import zipfile

    if dir is None:
        dir = "./"

    # confirm dir/example_data does not exist
    if os.path.exists(os.path.join(dir, "example_data")):
        admix.logger.info(
            f"Example data set already exists at {dir}/example_data, skip downloading"
        )
    else:
        urllib.request.urlretrieve(
            "https://github.com/KangchengHou/admix-kit/releases/download/v0.1.2/example_data.zip",
            os.path.join(dir, "example_data.zip"),
        )
        with zipfile.ZipFile(os.path.join(dir, "example_data.zip"), "r") as f:
            f.extractall(dir)

        admix.logger.info(f"Example data set downloaded to {dir}/example_data")

        # remove the zip file
        os.remove(os.path.join(dir, "example_data.zip"))


def load_toy_admix() -> Dataset:
    """
    Load toy admixed data set with African and European ancestries

    Returns
    -------
    Dataset
    """
    dset = admix.io.read_dataset(join(get_test_data_dir(), "toy-admix"), n_anc=2)
    return dset


def load_toy() -> List[Dataset]:
    """Load toy dataset

    Load simulated
    (1) 50 admixed individuals
    (2) 50 EUR individuals
    (3) 50 AFR individuals

    5000 SNPs

    Returns
    -------
    List[admix.Dataset]
        [dset_admix, dset_eur, dset_afr]
    """

    # TODO: change the data format, use .pgen and .lanc
    import xarray as xr

    module_path = dirname(__file__)
    test_data_path = join(module_path, "../../tests/test-data")
    dset_eur = xr.open_zarr(join(test_data_path, "eur.zip"))
    dset_afr = xr.open_zarr(join(test_data_path, "afr.zip"))
    dset_admix = xr.open_zarr(join(test_data_path, "admix.zip"))

    dset_list = [
        Dataset(
            geno=np.swapaxes(dset_admix.geno.data, 0, 1),
            lanc=np.swapaxes(dset_admix.lanc.data, 0, 1),
            n_anc=2,
            indiv=dset_admix.indiv.to_dataframe().drop(columns=["indiv"]),
            snp=dset_admix.snp.to_dataframe().drop(columns=["snp"]),
        )
    ]
    for dset in [dset_eur, dset_afr]:
        dset_list.append(
            Dataset(
                geno=np.swapaxes(dset.geno.data, 0, 1),
                n_anc=1,
                indiv=dset.indiv.to_dataframe().drop(columns=["indiv"]),
                snp=dset.snp.to_dataframe().drop(columns=["snp"]),
            ),
        )

    return dset_list


"""
load compiled data sets
"""


# PAGE admixed individuals with EUR - AFR ancestries in hapmap3 density
# def load_page_eur_afr_hm3(
#     chrom=None,
#     GENO_DIR="/u/project/pasaniuc/pasaniucdata/admixture/projects/PAGE-QC/s03_aframr/dataset/hm3.zarr/",
#     PHENO_DIR="/u/project/sgss/PAGE/phenotype/",
# ):
#     if chrom is None:
#         chrom = np.arange(1, 23)
#     elif isinstance(chrom, List):
#         chrom = np.array(chrom)
#     elif isinstance(chrom, int):
#         chrom = np.array([chrom])
#     else:
#         raise ValueError("chrom must be None, List or int")

#     trait_cols = [
#         # Inflammtory traits
#         "crp",
#         "total_wbc_cnt",
#         "mean_corp_hgb_conc",
#         "platelet_cnt",
#         # lipid traits
#         "hdl",
#         "ldl",
#         "triglycerides",
#         "total_cholesterol",
#         # lifestyle traits
#         "cigs_per_day_excl_nonsmk_updated",
#         "coffee_cup_day",
#         # glycemic traits
#         "a1c",
#         "insulin",
#         "glucose",
#         "t2d_status",
#         # electrocardiogram traits
#         "qt_interval",
#         "qrs_interval",
#         "pr_interval",
#         # blood pressure traits
#         "systolic_bp",
#         "diastolic_bp",
#         "hypertension",
#         # anthropometric traits
#         "waist_hip_ratio",
#         "height",
#         "bmi",
#         # kidney traits
#         "egfrckdepi",
#     ]

#     covar_cols = ["study", "age", "sex", "race_ethnicity", "center"] + [
#         f"geno_EV{i}" for i in range(1, 51)
#     ]

#     race_encoding = {
#         1: "Unclassified",
#         2: "African American",
#         3: "Hispanic/Latino",
#         4: "Asian",
#         5: "Native Hawaiian",
#         6: "Native American",
#         7: "Other",
#     }

#     race_color = {
#         "Unclassified": "#ffffff",
#         "African American": "#e9d16a",
#         "Hispanic/Latino": "#9a3525",
#         "Asian": "#3c859d",
#         "Native Hawaiian": "#959f6e",
#         "Native American": "#546f7b",
#         "Other": "#d07641",
#     }

#     df_pheno = pd.read_csv(
#         join(PHENO_DIR, "MEGA_page-harmonized-phenotypes-pca-freeze2-2016-12-14.txt"),
#         sep="\t",
#         na_values=".",
#         low_memory=False,
#     )
#     df_pheno["race_ethnicity"] = df_pheno["race_ethnicity"].map(race_encoding)

#     dset_list = []
#     for i_chr in tqdm(chrom):
#         dset_list.append(xr.open_zarr(join(GENO_DIR, f"chr{i_chr}.zip")))

#     dset = xr.concat(dset_list, dim="snp")

#     df_aframr_pheno = df_pheno.set_index("PAGE_Subject_ID").loc[
#         dset.indiv.values, trait_cols + covar_cols
#     ]

#     for col in df_aframr_pheno.columns:
#         dset[f"{col}@indiv"] = ("indiv", df_aframr_pheno[col].values)

#     for col in ["center", "study", "race_ethnicity"]:
#         dset[f"{col}@indiv"] = dset[f"{col}@indiv"].astype(str)

#     # format the dataset to follow the new standards
#     for k in dset.data_vars.keys():
#         if k.endswith("@indiv"):
#             dset.coords[k.split("@")[0]] = ("indiv", dset.data_vars[k].data)
#         if k.endswith("@snp"):
#             dset.coords[k.split("@")[0]] = ("snp", dset.data_vars[k].data)
#     dset = dset.drop_vars(
#         [
#             k
#             for k in dset.data_vars.keys()
#             if (k.endswith("@indiv") or k.endswith("@snp"))
#         ]
#     )

#     dset = dset.rename({n: n.split("@")[0] for n in [k for k in dset.coords.keys()]})

#     return dset


# def load_page_eur_afr_imputed(
#     region: str,
#     imputed_vcf_dir: str = "/u/project/pasaniuc/pasaniucdata/admixture/projects/PAGE-QC/s01_dataset/all",
# ):
#     """
#     PAGE admixed individuals with EUR - AFR ancestries in imputed density

#     Parameters
#     ----------
#     region : str
#         Region name, in the format of "chrom:start-stop", e.g. "1:1-100000"
#     imputed_vcf_dir : str
#         Directory of imputed VCF files
#     """
#     chrom = int(region.split(":")[0])
#     start, stop = [int(i) for i in region.split(":")[1].split("-")]
#     print(chrom)
#     # load hm3 data set
#     dset_hm3 = load_page_eur_afr_hm3(chrom=chrom)

#     dset = admix.io.read_vcf(
#         join(imputed_vcf_dir, f"chr{chrom}.vcf.gz"),
#         region=f"{chrom}:{start}-{stop}",
#     ).sel(indiv=dset_hm3.indiv.values)

#     dset = admix.data.impute_lanc(dset=dset, dset_ref=dset_hm3)
#     dset.attrs["n_anc"] = 2
#     return dset


# UK Biobank admixed individuals with EUR - AFR ancestries in hapmap3 density
def load_ukb_eur_afr_hm3(
    chrom=None,
    GENO_DIR="/u/project/pasaniuc/pasaniucdata/admixture/projects/admix-prs-uncertainty/data/admix-analysis/dataset",
    PHENO_DIR="/u/project/pasaniuc/pasaniucdata/admixture/projects/admix-prs-uncertainty/data/pheno",
):
    if chrom is None:
        chrom = np.arange(1, 23)
    elif isinstance(chrom, List):
        chrom = np.array(chrom)
    elif isinstance(chrom, int):
        chrom = np.array([chrom])
    else:
        raise ValueError("chrom must be None, List or int")

    dset_list = []
    for i_chr in tqdm(chrom):
        dset_list.append(xr.open_zarr(join(GENO_DIR, f"chr{i_chr}.zarr")))

    dset = xr.concat(dset_list, dim="snp")

    trait_cols = ["height"]

    for trait in trait_cols:
        # raw phenotype
        df_pheno = pd.read_csv(
            join(PHENO_DIR, f"admix.{trait}.pheno"),
            delim_whitespace=True,
            low_memory=False,
        )

        df_pheno.index = df_pheno.FID.astype(str) + "_" + df_pheno.IID.astype(str)
        df_pheno = df_pheno.drop(columns=["FID", "IID"])
        dset.coords[trait] = "indiv", df_pheno[trait].reindex(dset.indiv.values)

        # residual phenotype
        df_pheno = pd.read_csv(
            join(PHENO_DIR, f"admix.{trait}.residual_pheno"),
            delim_whitespace=True,
            low_memory=False,
        )

        df_pheno.index = df_pheno.FID.astype(str) + "_" + df_pheno.IID.astype(str)
        df_pheno = df_pheno.drop(columns=["FID", "IID"])
        dset.coords[f"{trait}_residual"] = (
            "indiv",
            df_pheno[f"{trait}-residual"].reindex(dset.indiv.values),
        )

    df_covar = pd.read_csv(
        join(PHENO_DIR, f"admix.covar"),
        delim_whitespace=True,
        low_memory=False,
    )

    df_covar.index = df_covar.FID.astype(str) + "_" + df_covar.IID.astype(str)
    df_covar = df_covar.drop(columns=["FID", "IID"])

    for col in df_covar:
        dset.coords[col] = "indiv", df_covar[col].reindex(dset.indiv.values)

    dset = dset.rename({n: n.split("@")[0] for n in [k for k in dset.coords.keys()]})

    return Dataset(
        geno=np.swapaxes(dset.geno.data, 0, 1),
        lanc=np.swapaxes(dset.lanc.data, 0, 1),
        n_anc=2,
        indiv=dset.indiv.to_dataframe().drop(columns=["indiv"]),
        snp=dset.snp.to_dataframe().drop(columns=["snp"]),
    )


def load_ukb_eur_afr_imputed(
    region: str,
    imputed_vcf_dir: str = "/u/project/pasaniuc/pasaniucdata/admixture/projects/admix-prs-uncertainty/data/PLINK/admix/topmed",
):
    """
    UKB admixed individuals with EUR - AFR ancestries in imputed density

    Parameters
    ----------
    region : str
        Region name, in the format of "chrom:start-stop", e.g. "1:1-100000"
    imputed_vcf_dir : str
        Directory of imputed VCF files
    """
    chrom = int(region.split(":")[0])
    start, stop = [int(i) for i in region.split(":")[1].split("-")]
    # load hm3 data set
    dset_hm3 = load_ukb_eur_afr_hm3(chrom=chrom)

    dset = admix.io.read_vcf(
        join(imputed_vcf_dir, f"chr{chrom}/chr{chrom}.sample.imputed.vcf.gz"),
        region=f"chr{chrom}:{start}-{stop}",
    )
    if dset is None:
        warnings.warn(f"No SNPs found in {region}, `None` is returned")
        return None

    dset = dset.sel(indiv=dset_hm3.indiv.values)

    # impute local ancestry
    dset = admix.data.impute_lanc(dset=dset, dset_ref=dset_hm3)
    dset.attrs["n_anc"] = 2
    return dset


def load_lab_dataset(name: str, chrom: int = None, region: str = None) -> xr.Dataset:
    """Load prepared dataset in Bogdan lab, currently available
    if you use this function on cluster, and have the proper access to the data.
    Contact kangchenghou@gmail.com before using this function.

    Parameters
    ----------
    name : str
        Name of the dataset
        - `simulate_eur_afr.20_80`: Simulated admixture of 20% EUR and 80% AFR
        - `simulate_eur_afr.50_50`: Simulated admixture of 50% EUR and 50% AFR
        - `ukb_eur_afr_hm3`: Admixed individuals of European and African ancestries
            in UK Biobank with hapmap3 SNPs
        - `ukb_eur_afr_imputed`: Admixed individuals of European and African ancestries
            in UK Biobank with imputed SNPs
        - `page_eur_afr_hm3`: Admixed individuals of European and African ancestries
            in PAGE with hapmap3 SNPs
        - `page_eur_afr_imputed`: Admixed individuals of European and African ancestries
            in PAGE with imputed SNPs
    chrom : int, optional
        Chromosome number
    region : str, optional
        Region name

    Returns
    -------
    xr.Dataset
        dataset
    """
    if name.startswith("simulate_eur_afr"):
        assert name in [
            "simulate_eur_afr.20_80",
            "simulate_eur_afr.50_50",
        ]
        root_dir = "/u/project/pasaniuc/pasaniucdata/admixture/dataset"
        assert os.path.isdir(root_dir), f"check that you have access to {root_dir}"
        dset = xr.open_zarr(join(root_dir, name + ".zip"))
    elif name == "ukb_eur_afr_hm3":
        assert region is None, "region must not be specified when loading hm3 data"
        dset = load_ukb_eur_afr_hm3(chrom=chrom)
    elif name == "ukb_eur_afr_imputed":
        assert region is not None, "region must be specified when loading imputed data"
        assert chrom is None, "chrom must not be specified when loading imputed data"
        dset = load_ukb_eur_afr_imputed(region=region)
    else:
        raise NotImplementedError
    # elif name == "page_eur_afr_hm3":
    #     assert region is None, "region must not be specified when loading hm3 data"
    #     dset = load_page_eur_afr_hm3(chrom=chrom)
    # elif name == "page_eur_afr_imputed":
    #     assert region is not None, "region must be specified when loading imputed data"
    #     assert chrom is None, "chrom must not be specified when loading imputed data"
    #     dset = load_page_eur_afr_imputed(region=region)
    return dset



import warnings
import matplotlib.pyplot as plt
from matplotlib import collections as mc
from matplotlib import patheffects

import numpy as np
import pandas as pd
from scipy import stats

import admix
from admix.data import quantile_normalize
from admix.data import lambda_gc

from typing import Dict


def pca(
    df_pca: pd.DataFrame,
    x: str = "PC1",
    y: str = "PC2",
    label_col: str = None,
    label_order: list = None,
    s=5,
    legend_loc="on data",
    alpha=None,
    ax=None,
):
    """PCA plot

    Parameters
    ----------
    df_pca : pd.DataFrame
        dataframe with PCA components
    x : str, optional
        x-axis, by default "PC1"
    y : str, optional
        y-axis, by default "PC2"
    label_col : str, optional
        column name for labels, by default None
    s : float, optional
    """
    if alpha is None:
        alpha = 1.0
    else:
        assert isinstance(alpha, float) or isinstance(alpha, dict)
    if ax is None:
        ax = plt.gca()

    if label_order is None:
        label_order = df_pca[label_col].unique()

    ax.set_xlabel(x)
    ax.set_ylabel(y)

    if label_col is None:
        return

    # otherwise label_col is present
    for label in label_order:
        group = df_pca.loc[df_pca[label_col] == label, :]
        if isinstance(alpha, dict):
            label_alpha = alpha[label] if label in alpha else 1.0
        else:
            label_alpha = alpha
        ax.scatter(group[x], group[y], s=s, label=label, alpha=label_alpha)

    if legend_loc == "on data":

        all_pos = (
            pd.DataFrame(df_pca[[x, y, label_col]])
            .groupby(label_col, observed=True)
            .median()
            .sort_index()
        )

        for label, x_pos, y_pos in all_pos.itertuples():
            ax.text(
                x_pos,
                y_pos,
                label,
                # weight="bold",
                path_effects=[patheffects.withStroke(linewidth=2.5, foreground="w")],
                verticalalignment="center",
                horizontalalignment="center",
            )

    legend = ax.legend()
    for lh in legend.legendHandles:
        lh.set_alpha(1)
        lh.set_sizes([30])


def joint_pca(
    df_pc,
    x="PC1",
    y="PC2",
    sample_alpha=0.1,
    axes=None,
    figsize=(8.5, 4),
    label_col="SUPERPOP",
    sample_label="SAMPLE",
):
    """Joint PCA plot

    Parameters
    ----------
    df_pc : pd.DataFrame
        dataframe with PCA components
    eigenval : np.ndarray
        eigenvalues
    """
    new_axes = axes is None
    if new_axes:
        fig, axes = plt.subplots(figsize=figsize, dpi=150, ncols=2)

    admix.plot.pca(
        df_pc[df_pc[label_col] != sample_label],
        x=x,
        y=y,
        label_col=label_col,
        ax=axes[0],
    )
    assert set([x, y]).issubset(
        df_pc.columns
    ), f"{x} and {y} must be in the columns of df_pc"

    x_pos, y_pos = df_pc.columns.get_loc(x), df_pc.columns.get_loc(y)
    xlabel, ylabel = x, y
    axes[0].set_xlabel(xlabel)
    axes[0].set_ylabel(ylabel)

    admix.plot.pca(
        df_pc,
        x=x,
        y=y,
        label_col=label_col,
        alpha={sample_label: sample_alpha},
        ax=axes[1],
    )
    axes[1].set_xlabel(xlabel)
    axes[1].set_ylabel(ylabel)

    if new_axes:
        return fig, axes


def qq(pval, label=None, ax=None, bootstrap_ci=False):
    """qq plot of p-values

    Parameters
    ----------
    pval : np.ndarray
        p-values, array-like
    ax : matplotlib.axes, optional
        by default None
    return_lambda_gc : bool, optional
        whether to return the lambda GC, by default False
    """
    if ax is None:
        ax = plt.gca()

    pval = np.array(pval)
    pval = pval[~np.isnan(pval)]
    expected_pval = stats.norm.sf(quantile_normalize(-pval))
    ax.scatter(-np.log10(expected_pval), -np.log10(pval), s=2, label=label)
    lim = max(-np.log10(expected_pval))
    ax.plot([0, lim], [0, lim], "r--")
    ax.set_xlabel("Expected -$\log_{10}(p)$")
    ax.set_ylabel("Observed -$\log_{10}(p)$")
    if bootstrap_ci == True:
        lgc, lgc_ci = lambda_gc(pval, bootstrap_ci=True)
    else:
        lgc = lambda_gc(pval, bootstrap_ci=False)

    if bootstrap_ci:
        print(f"lambda GC: {lgc:.3g} [{lgc_ci[0]:.3g}, {lgc_ci[1]:.3g}]")
        return lgc, lgc_ci
    else:
        print(f"lambda GC: {lgc:.3g}")
        return lgc


def manhattan(
    pval,
    chrom=None,
    pos=None,
    axh_y=-np.log10(5e-8),
    s=0.1,
    label=None,
    ax=None,
    color="#3b76af",
):
    """Manhatton plot of p-values

    Parameters
    ----------
    chrom : np.ndarray
        array-like
    pval : np.ndarray
        p-values, array-like
    pos: np.ndarray
        array-like, position for each SNP, if provided, position will be used
    axh_y : np.ndarray, optional
        horizontal line for genome-wide significance, by default -np.log10(5e-8)
    s : float, optional
        dot size, by default 0.1
    ax : matplotlib.axes, optional
        axes, by default None
    """

    if ax is None:
        ax = plt.gca()

    assert (chrom is None) or (pos is None), "chrom and pos cannot be both provided"
    if pos is None:
        pos_provided = False
        pos = np.arange(len(pval))
    else:
        pos_provided = True
        assert len(pos) == len(pval)

    if chrom is None:
        # use snp index
        if pos_provided:
            ax.scatter(
                pos / 1e6,
                -np.log10(pval),
                s=s,
                label=label,
                facecolor=color,
                marker="o",
            )
            ax.set_xlabel("SNP position (Mb)")
        else:
            ax.scatter(pos, -np.log10(pval), s=s, label=label, c=color)
            ax.set_xlabel("SNP index")

    else:
        assert pos_provided is False
        assert len(chrom) == len(pval)
        color_list = ["#1b9e77", "#d95f02"]
        # plot dots for odd and even chromosomes
        for mod in range(2):
            index = np.where(chrom % 2 == mod)[0]
            ax.scatter(
                pos[index],
                -np.log10(pval)[index],
                s=s,
                color=color_list[mod],
                label=label,
            )

        # label unique chromosomes
        xticks = []
        xticklabels = []
        for chrom_i in np.unique(chrom):
            xticks.append(np.where(chrom == chrom_i)[0].mean())
            xticklabels.append(chrom_i)

        ax.set_xticks(xticks)
        ax.set_xticklabels(xticklabels, rotation=90, fontsize=8)
        ax.set_xlabel("Chromosome")

    ax.set_ylabel("-$\log_{10}(P)$")
    if axh_y is not None:
        ax.axhline(y=axh_y, color="r", ls="--")


def susie(pip, dict_cs, pos=None, ax=None, cmap="Set1"):
    cmap = plt.get_cmap(cmap)
    if ax is None:
        ax = plt.gca()
    if pos is None:
        pos = np.arange(len(pip))
        pos_provided = False
    else:
        pos_provided = True
        assert len(pos) == len(pip)
        pos = pos / 1e6

    ax.scatter(x=pos, y=pip, s=3, color="gray")

    for i, cs in enumerate(dict_cs):
        cs_pos = dict_cs[cs]
        ax.scatter(
            x=pos[cs_pos],
            y=pip[cs_pos],
            s=15,
            edgecolors=cmap.colors[i],
            facecolors=cmap.colors[i],
            alpha=0.6,
        )
    if pos_provided:
        ax.set_xlabel("SNP position (Mb)")
    else:
        ax.set_xlabel("SNP index")
    ax.set_ylabel("PIP")
    ax.set_ylim(-0.05, 1.05)


def lanc(
    dset: admix.Dataset = None,
    lanc: np.ndarray = None,
    ax=None,
    max_indiv: int = None,
) -> None:
    """
    Plot local ancestry.

    Parameters
    ----------
    dset: xarray.Dataset
        A dataset containing the local ancestry matrix.
    lanc: np.ndarray
        A numpy array of shape (n_snp, n_indiv, 2)
    ax: matplotlib.Axes
        A matplotlib axes object to plot on. If None, will create a new one.
    max_indiv: int
        The maximum number of individuals to plot.
        If None, will plot the first 10 individuals
    Returns
    -------
    ax: matplotlib.Axes
    """
    # if dataset is provided, use it to extract lanc
    if dset is not None:
        lanc = dset.lanc.compute()
        pos = dset.snp.POS.values
        BP_POS = True
    else:
        assert lanc is not None, "either dataset or lanc must be provided"
        pos = np.arange(lanc.shape[0])
        BP_POS = False
    # append dummy snp at the end to make plotting easier
    pos = np.concatenate([pos, [pos[-1] + 1]])

    assert lanc.shape[2] == 2, "lanc must be of shape (n_snp, n_indiv, 2)"
    n_snp, n_indiv = lanc.shape[0:2]

    if max_indiv is not None:
        n_plot_indiv = min(max_indiv, n_indiv)
    else:
        n_plot_indiv = min(n_indiv, 10)
        if n_plot_indiv < n_indiv:
            warnings.warn(
                f"Only the first {n_plot_indiv} are plotted. To plot more individuals, increase `max_indiv`"
            )
    if ax is None:
        ax = plt.gca()

    start = []
    stop = []
    label = []
    row = []

    for i_indiv in range(n_plot_indiv):
        for i_ploidy in range(2):
            a = lanc[:, i_indiv, i_ploidy]
            switch = np.where(a[1:] != a[0:-1])[0]
            switch = np.concatenate([[0], switch, [len(a)]])
            for i_switch in range(len(switch) - 1):
                start_idx, stop_idx = switch[i_switch], switch[i_switch + 1]
                if BP_POS:
                    start.append(pos[start_idx] / 1e6)
                    stop.append(pos[stop_idx] / 1e6)
                else:
                    start.append(start_idx)
                    stop.append(stop_idx)

                label.append(a[start_idx + 1])
                row.append(i_indiv - 0.1 + i_ploidy * 0.2)

    df_plot = pd.DataFrame({"start": start, "stop": stop, "label": label, "row": row})

    lines = [[(r.start, r.row), (r.stop, r.row)] for _, r in df_plot.iterrows()]

    cmap = plt.get_cmap("tab10")

    for i, (label, group) in enumerate(df_plot.groupby("label")):

        lc = mc.LineCollection(
            [lines[l_i] for l_i in group.index],
            linewidths=2,
            label=label,
            color=cmap(i),
        )
        ax.add_collection(lc)

    ax.legend()
    ax.autoscale()

    if BP_POS:
        ax.set_xlabel("SNP position (Mb)")
    else:
        ax.set_xlabel("SNP index")
    ax.set_ylabel("Individuals")
    ax.set_yticks([])
    ax.set_yticklabels([])


def admixture(
    a: np.ndarray,
    labels=None,
    label_orders=None,
    ax=None,
) -> None:
    """
    Plot admixture.

    Parameters
    ----------
    a: np.ndarray
        A numpy array of shape (n_indiv, n_snp, 2)
    labels: list
        A list of labels for each individual.
    label_orders: list
        A list of orderings for the individuals.
    ax: matplotlib.Axes
        A matplotlib axes object to plot on. If None, will create a new one.

    Returns
    -------
    None
    """

    n_indiv, n_pop = a.shape

    # reorder based on labels
    if labels is not None:
        dict_label_range = dict()
        reordered_a = []
        unique_labels = np.unique(labels)

        if label_orders is not None:
            assert set(label_orders) == set(
                unique_labels
            ), "label_orders must cover all unique labels"
            unique_labels = label_orders
        cumsum = 0
        for label in unique_labels:
            reordered_a.append(a[labels == label, :])
            dict_label_range[label] = [cumsum, cumsum + sum(labels == label)]
            cumsum += sum(labels == label)
        a = np.vstack(reordered_a)

    if ax is None:
        ax = plt.gca()

    cmap = plt.get_cmap("tab10")
    bottom = np.zeros(n_indiv)

    for i_pop in range(n_pop):
        ax.bar(
            np.arange(n_indiv),
            height=a[:, i_pop],
            width=1,
            bottom=bottom,
            facecolor=cmap(i_pop),
            edgecolor=cmap(i_pop),
        )
        bottom += a[:, i_pop]

    ax.tick_params(axis="both", left=False, labelleft=False)

    if labels is not None:
        seps = sorted(np.unique(np.concatenate([r for r in dict_label_range.values()])))
        for x in seps[1:-1]:
            ax.vlines(x - 0.5, ymin=0, ymax=1, color="black")

        ax.set_xticks([np.mean(dict_label_range[label]) for label in dict_label_range])
        ax.set_xticklabels([label for label in dict_label_range])
    else:
        ax.get_xaxis().set_ticks([])
    for pos in ["top", "right", "bottom", "left"]:
        ax.spines[pos].set_visible(False)
    return ax


def compare_pval(
    x_pval: np.ndarray,
    y_pval: np.ndarray,
    xlabel: str = None,
    ylabel: str = None,
    ax=None,
    s: int = 5,
):
    """Compare two p-values.

    Parameters
    ----------
    x_pval: np.ndarray
        The p-value for the first variable.
    y_pval: np.ndarray
        The p-value for the second variable.
    xlabel: str
        The label for the first variable.
    ylabel: str
        The label for the second variable.
    ax: matplotlib.Axes
        A matplotlib axes object to plot on. If None, will create a new one.
    """
    if ax is None:
        ax = plt.gca()
    if not isinstance(x_pval, np.ndarray):
        x_pval = np.array(x_pval)
    if not isinstance(y_pval, np.ndarray):
        y_pval = np.array(y_pval)
    nonnan_idx = ~np.isnan(x_pval) & ~np.isnan(y_pval)
    x_pval, y_pval = -np.log10(x_pval[nonnan_idx]), -np.log10(y_pval[nonnan_idx])
    ax.scatter(x_pval, y_pval, s=s)
    lim = max(np.nanmax(x_pval), np.nanmax(y_pval)) * 1.1
    ax.plot([0, lim], [0, lim], "k--", alpha=0.5, lw=1, label="y=x")

    # add a regression line
    slope = np.linalg.lstsq(x_pval[:, None], y_pval[:, None], rcond=None)[0].item()

    ax.axline(
        (0, 0),
        slope=slope,
        color="black",
        ls="--",
        lw=1,
        label=f"y={slope:.2f} x",
    )
    ax.legend()
    if xlabel is not None:
        ax.set_xlabel(xlabel)
    if ylabel is not None:
        ax.set_ylabel(ylabel)


def rg_posterior(
    xs: np.ndarray,
    dict_loglik: Dict[str, np.ndarray],
    ci=[0.5, 0.95],
    s=11,
    colors="black",
    markers="o",
    ax=None,
):
    """
    Plot the posterior distribution

    Parameters
    ----------
    xs: np.ndarray
        list of x coordinates
    dict_loglik: Dict[np.ndarray]
        trait -> list of log-likelihoods
    ci: Union[float, List[float]]
        ci to plot, can be 1 float or two float
    colors:
        ["darkblue"] * (len(est) - 1) + ["darkred"]
    markers:
        ["o"] * (len(est) - 1) + ["^"]
    """
    if ax is None:
        ax = plt.gca()

    assert len(ci) == 2, "Currently must plot 2 CIs"
    assert ci[0] < ci[1], "Smaller CI should come first"
    assert np.all([len(xs) == len(dict_loglik[t]) for t in dict_loglik])

    trait_list = list(dict_loglik.keys())[::-1]
    if isinstance(colors, list):
        colors = colors[::-1]
    elif isinstance(colors, str):
        colors = [colors] * len(trait_list)
    if isinstance(markers, list):
        markers = markers[::-1]
    elif isinstance(markers, str):
        markers = [markers] * len(trait_list)

    dict_mode = {trait: xs[dict_loglik[trait].argmax()] for trait in trait_list}

    dict_ci_err: Dict[int, Dict] = {ci[0]: dict(), ci[1]: dict()}

    for trait in trait_list:
        mode = dict_mode[trait]
        for each_ci in ci:
            hdi = admix.data.hdi(xs, dict_loglik[trait], ci=each_ci)
            assert not isinstance(
                hdi, list
            ), f"HPDI for {trait} contains multiple intervals {hdi}, indicating lack of data. Please rerun this function after remove this trait."
            dict_ci_err[each_ci][trait] = [mode - hdi[0], hdi[1] - mode]

    mode = np.array([dict_mode[trait] for trait in trait_list])
    lw_list = [2.5, 1.0]
    for i, each_ci in enumerate(ci):
        ci_low = [dict_ci_err[each_ci][trait][0] for trait in trait_list]
        ci_high = [dict_ci_err[each_ci][trait][1] for trait in trait_list]
        ax.errorbar(
            y=np.arange(len(mode)),
            x=mode,
            xerr=(ci_low, ci_high),
            fmt=" ",
            lw=lw_list[i],
            ecolor=colors,
        )
        if i == 0:
            for j, trait in enumerate(trait_list):
                ax.scatter(
                    x=mode[j],
                    y=j,
                    marker=markers[j],
                    color=colors[j],
                    s=s,
                )

    for y in np.arange(len(mode)):
        ax.axhline(y=y, color="gray", ls="dotted", lw=0.5, alpha=0.8)

    ax.set_xlim(0, 1.1)
    ax.set_xlabel("Highest probability density of $r_{admix}$")
    ax.set_yticks(np.arange(len(mode)))
    ax.set_ylim(-0.5, len(mode) - 0.5)
    ax.set_yticklabels(
        trait_list,
        fontsize=8,
    )

    # annotation
    ax.tick_params(left=False, pad=-1)
    ax.axvline(x=1.0, color="red", ls="--", lw=0.8, alpha=0.4)
    ax.set_title("Estimated $r_{admix}$", fontsize=10, x=0.5)



import admix
import subprocess
import dapgen
import os
import glob
import numpy as np
import pandas as pd
from ._utils import log_params


def lanc(
    pfile: str,
    ref_pfile: str,
    ref_pop_col: str,
    ref_pops: str,
    out: str,
):
    log_params("lanc", locals())

    sample_dset = admix.io.read_dataset(pfile=pfile)
    ref_dset = admix.io.read_dataset(pfile=ref_pfile)

    assert set(sample_dset.snp.index) == set(ref_dset.snp.index), (
        "`pfile` and `ref_pfile` must have the same snp index"
        "(snp match feature coming soon)."
    )

    ref_dsets = [
        ref_dset[:, (ref_dset.indiv[ref_pop_col] == pop).values] for pop in ref_pops
    ]
    est = admix.ancestry.lanc(sample_dset=sample_dset, ref_dsets=ref_dsets)
    admix.data.Lanc(array=est).write(out)


def lanc_count(lanc: str, out: str, n_anc: int = None):
    """Count the number / proportion of local ancestries for each individual

    Parameters
    ----------
    lanc : str
        path to the lanc file, this can be a .lanc file, a wildcard of .lanc files,
        or a directory containing .lanc files. If the corresponding .psam file is
        present, the .psam file will be used as the individual list.
    out : str
        path to the output file
    n_anc : int
        number of ancestral populations in the data
    """
    log_params("lanc-count", locals())
    if lanc.endswith(".lanc"):
        lanc_path = [lanc]
    elif "*" in lanc:
        lanc_path = glob.glob(lanc)
    elif os.path.isdir(lanc):
        lanc_path = [p for p in glob.glob(lanc + "/*.lanc")]
    else:
        raise ValueError("Unable to parse lanc pathname")

    admix.logger.info(f"Found {len(lanc_path)} lanc files: {','.join(lanc_path)}")
    # read psam if available
    psam_path = [p.replace(".lanc", ".psam") for p in lanc_path]

    if all(os.path.exists(p) for p in psam_path):
        # check all psam files have the same individual ID
        psam_indiv = [dapgen.read_psam(p).index for p in psam_path]
        assert all(
            psam_indiv[0].equals(i) for i in psam_indiv[1:]
        ), "Individuals in psam files do not match"
        indiv_list = psam_indiv[0].values
    elif not any(os.path.exists(p) for p in psam_path):
        indiv_list = None
    else:
        raise ValueError("either .psam all exists or none exists")

    lanc_mat = admix.data.Lanc(lanc_path[0])
    n_indiv = lanc_mat.n_indiv
    if indiv_list is not None:
        assert n_indiv == len(
            indiv_list
        ), "Number of individuals in lanc and psam files do not match"
    else:
        indiv_list = np.arange(n_indiv).astype(str)

    lanc_count = lanc_mat.lanc_count()
    if n_anc is not None:
        assert (
            lanc_count.shape[1] == n_anc
        ), "Number of ancestral populations do not match"
    else:
        n_anc = lanc_count.shape[1]
        admix.logger.info(f"Inferred number of ancestral populations: {n_anc}")

    for p in lanc_path[1:]:
        lanc_count += admix.data.Lanc(p).lanc_count(n_anc=n_anc)

    lanc_prop = lanc_count / lanc_count.sum(axis=1, keepdims=True)
    admix.logger.info(f"Writing lanc count file: {out}")

    count_cols = [f"COUNT{i + 1}" for i in range(n_anc)]
    prop_cols = [f"PROP{i+1}" for i in range(n_anc)]
    df_res = pd.DataFrame(
        data=np.concatenate([lanc_count, lanc_prop], axis=1),
        index=indiv_list,
        columns=count_cols + prop_cols,
    )

    df_res[count_cols] = df_res[count_cols].astype(int)
    df_res.index.name = "indiv"
    df_res.to_csv(out, sep="\t", float_format="%.4g")


def lanc_convert(pfile: str, out: str, rfmix: str = None, raw: str = None):
    """Convert local ancestry inference results (e.g. RFmix .msp.tsv) to a .lanc file

    Parameters
    ----------
    pfile : str
        Path to the pfile. The path is without the .pgen suffix
    out : str
        Path to the output file
    rfmix : str
        Path to the rfmix .msp.tsv file,
    raw : str
        Path to the raw file
    """
    log_params("lanc-convert", locals())

    # only one of rfmix and raw should be specified
    assert (rfmix is None) + (
        raw is None
    ) == 1, "Only one of rfmix and raw should be specified"
    if rfmix is not None:
        geno, df_snp, df_indiv = dapgen.read_pfile(pfile, phase=True)
        admix.logger.info(f"Reading rfmix file: {rfmix}")
        lanc = admix.io.read_rfmix(
            path=rfmix,
            df_snp=df_snp,
            df_indiv=df_indiv,
        )
        admix.logger.info(f"Obtaining local ancestry {lanc}")
        admix.logger.info(f"Writing lanc file: {out}")
        lanc.write(out)

    if raw is not None:
        assert False, "raw not implemented yet"


def lanc_impute(pfile: str, ref_pfile: str, out: str = None):
    """Impute the local ancestry for `pfile` using `ref_pfile`

    Parameters
    ----------
    pfile : str
        Path to the pfile
    ref_pfile : str
        Path to the reference pfile
    out : str
        Path to the output pfile (default to pfile + ".lanc")
    """
    log_params("lanc-impute", locals())

    # check <pfile>.lanc does not exist
    assert not os.path.exists(pfile + ".lanc"), "`pfile` already has a .lanc file"

    sample_dset = admix.io.read_dataset(pfile=pfile)
    ref_dset = admix.io.read_dataset(pfile=ref_pfile)
    ref_lanc = admix.data.Lanc(ref_pfile + ".lanc")

    sample_lanc = ref_lanc.impute(
        ref_dset.snp[["CHROM", "POS"]].values, sample_dset.snp[["CHROM", "POS"]].values
    )
    if out is None:
        out = pfile + ".lanc"
    assert not os.path.exists(out), f"out={out} already exists"
    sample_lanc.write(out)



#!/usr/bin/env python

import fire
from ._assoc import assoc
from ._geno import append_snp_info, calc_pgs, calc_partial_pgs, grm
from ._utils import log_params, get_1kg_ref, select_admix_indiv
from ._simulate import simulate_admix_pheno, simulate_pheno
from ._lanc import lanc, lanc_convert, lanc_count
from ._ext import (
    prune,
    pca,
    liftover,
    pfile_align_snp,
    pfile_merge_indiv,
    pfile_freq_within,
    subset_hapmap3,
    subset_pop_indiv,
    hapgen2,
    admix_simu,
    haptools_simu_admix,
    download_dependency,
)
from ._plot import plot_joint_pca
from ._genet_cor import (
    admix_grm,
    admix_grm_merge,
    genet_cor,
    admix_grm_rho,
    estimate_genetic_cor,
    summarize_genet_cor,
    meta_analyze_genet_cor,
)


def cli():
    """
    Entry point for the admix command line interface.
    """
    fire.Fire()


if __name__ == "__main__":
    fire.Fire()
